class MS2PIPError(Exception):
    pass


class InvalidPeptideError(MS2PIPError):
    pass


class UnknownModificationError(ValueError):
    pass


class InvalidPEPRECError(Exception):
    pass


class NoValidPeptideSequencesError(Exception):
    pass


class UnknownOutputFormatError(ValueError):
    pass


class UnknownFragmentationMethodError(ValueError):
    pass


class MissingConfigurationError(Exception):
    pass


class FragmentationModelRequiredError(Exception):
    pass


class InvalidModificationFormattingError(Exception):
    pass


class InvalidAminoAcidError(Exception):
    pass


class UnsupportedSpectrumFiletypeError(Exception):
    pass


class InvalidSpectrumError(Exception):
    pass


class EmptySpectrumError(InvalidSpectrumError):
    pass


class NoMatchingSpectraFound(MS2PIPError):
    pass

class TitlePatternError(MS2PIPError):
    pass

class InvalidXGBoostModelError(MS2PIPError):
    pass



"""
Create a spectral library starting from a proteome in fasta format.

The script runs through the following steps:
- In silico cleavage of proteins from the fasta file
- Remove peptide redundancy
- Add all variations of variable modifications (max 7 PTMs/peptide)
- Add variations on charge state
- Predict spectra with MS2PIP
- Write to various output file formats
"""


__author__ = "Ralf Gabriels"
__copyright__ = "CompOmics"
__credits__ = ["Ralf Gabriels", "Sven Degroeve", "Lennart Martens"]
__license__ = "Apache License, Version 2.0"
__email__ = "Ralf.Gabriels@ugent.be"


# Native libraries
import argparse
import json
import logging
import multiprocessing
from itertools import product
from math import ceil

# Third party libraries
import numpy as np
import pandas as pd
from Bio import SeqIO
from pyteomics.parser import cleave

from ms2pip.ms2pip_tools import spectrum_output
from ms2pip.ms2pip_tools.get_elude_predictions import get_elude_predictions

# MS2PIP
from ms2pip.ms2pipC import MS2PIP
from ms2pip.retention_time import RetentionTime


def ArgParse():
    parser = argparse.ArgumentParser(
        description="Create an MS2PIP-predicted spectral library, starting from a fasta file."
    )
    parser.add_argument(
        "fasta_filename",
        action="store",
        help="Path to the fasta file containing protein sequences",
    )
    parser.add_argument(
        "-o",
        dest="output_filename",
        action="store",
        help="Name for output file(s) (if not given, derived from input file)",
    )
    parser.add_argument(
        "-c",
        dest="config_filename",
        action="store",
        help="Name of configuration json file (default: fasta2speclib_config.json)",
    )

    args = parser.parse_args()
    return args


def get_params():
    args = ArgParse()

    if not args.config_filename:
        config_filename = "fasta2speclib_config.json"
    else:
        config_filename = args.config_filename

    with open(config_filename, "rt") as config_file:
        params = json.load(config_file)

    params.update(
        {
            "fasta_filename": args.fasta_filename,
            "log_level": logging.INFO,
        }
    )

    if args.output_filename:
        params["output_filename"] = args.output_filename
    else:
        params["output_filename"] = "_".join(
            params["fasta_filename"].split("\\")[-1].split(".")[:-1]
        )

    if not params["num_cpu"]:
        params["num_cpu"] = multiprocessing.cpu_count()

    return params


def prot_to_peprec(protein):
    """Cleave protein and return pd.DataFrame with valid peptides."""

    def validate_peptide(peptide, min_length, max_length):
        """Validate peptide by length and amino acids."""
        peplen = len(peptide)
        return (
            (peplen >= min_length)
            and (peplen <= max_length)
            and not any(aa in peptide for aa in ["B", "J", "O", "U", "X", "Z"])
        )

    params = get_params()

    pep_count = 0
    spec_ids = []
    peptides = []

    for peptide in cleave(
        str(protein.seq), params["cleavage_rule"], params["missed_cleavages"]
    ):
        pep_count += 1
        if validate_peptide(peptide, params["min_peplen"], params["max_peplen"]):
            spec_ids.append("{}_{:03d}".format(protein.id, pep_count))
            peptides.append(peptide)

    return pd.DataFrame(
        {
            "spec_id": spec_ids,
            "peptide": peptides,
            "modifications": "-",
            "charge": np.nan,
        }
    )


def get_protein_list(df):
    peptide_to_prot = {}
    for pi, pep in zip(df["spec_id"], df["peptide"]):
        pi = "_".join(pi.split("_")[0:2])
        if pep in peptide_to_prot.keys():
            peptide_to_prot[pep].append(pi)
        else:
            peptide_to_prot[pep] = [pi]
    df["protein_list"] = [list(set(peptide_to_prot[pep])) for pep in df["peptide"]]
    df = df[~df.duplicated(["peptide", "charge", "modifications"])]
    return df


def get_modifications_by_target(modifications):
    """Restructure modifications configuration to options per side chain or terminus."""
    mods_sidechain = {}
    mods_nterm = {}

    for mod in modifications:
        if mod["fixed"]:
            continue
        elif mod["n_term"]:
            if mod["amino_acid"]:
                if mod["amino_acid"] in mods_nterm:
                    mods_nterm[mod["amino_acid"]].append(mod["name"])
                else:
                    mods_nterm[mod["amino_acid"]] = [mod["name"]]
            else:
                if "any" in mods_nterm:
                    mods_nterm["any"].append(mod["name"])
                else:
                    mods_nterm["any"] = [mod["name"]]
        elif mod["amino_acid"]:
            if mod["amino_acid"] in mods_sidechain:
                mods_sidechain[mod["amino_acid"]].append(mod["name"])
            else:
                mods_sidechain[mod["amino_acid"]] = [mod["name"]]

    for aa, mods in mods_sidechain.items():
        mods.append(None)
    if "any" in mods_nterm:
        mods_nterm["any"].append(None)
    else:
        mods_nterm["any"] = [None]

    return mods_sidechain, mods_nterm


def get_modification_versions(
    peptide, modifications, mods_sidechain, mods_nterm, max_mods=3
):
    """Get MS²PIP modification strings for all potential versions."""
    possibilities_by_site = dict()

    # Assign possible modifications per residue (side chains)
    for pos, aa in enumerate(peptide):
        if aa in mods_sidechain:
            possibilities_by_site[pos + 1] = mods_sidechain[aa]

    # Assign possible modifications for N terminus
    if mods_nterm:
        possibilities_by_site[0] = []
    for site, mods in mods_nterm.items():
        if site == "any":
            possibilities_by_site[0].extend(mods)
        elif peptide[0] == site:
            possibilities_by_site[0].extend(mods)

    # Override with fixed modifications
    for mod in modifications:
        if not mod["fixed"]:
            continue
        if mod["n_term"]:
            if (mod["amino_acid"] and peptide[0] == mod["amino_acid"]) or not mod[
                "amino_acid"
            ]:
                possibilities_by_site[0] = [mod["name"]]
        elif mod["amino_acid"]:
            for pos, aa in enumerate(peptide):
                if aa == mod["amino_acid"]:
                    possibilities_by_site[pos + 1] = [mod["name"]]

    # Get all possible combinations of modifications for all sites
    mod_permutations = list(product(*possibilities_by_site.values()))
    mod_positions = possibilities_by_site.keys()

    # Get MS²PIP modifications strings for each combination
    mod_strings = []
    for perm in mod_permutations:
        mods = sorted(zip(mod_positions, perm))  # Zip permutations with positions
        mods = "|".join(
            f"{m[0]}|{m[1]}" for m in mods if m[1]
        )  # Make str for modified sites
        mod_strings.append(mods)

    # Filter by max modified sites (avoiding combinatorial explosion)
    mod_strings = list(
        filter(lambda x: (x.count("|") + 1) / 2 <= max_mods, mod_strings)
    )

    return mod_strings


def add_mods(tup):
    """
    See fasta2speclib_config.md for more information.
    """
    _, row = tup
    params = get_params()

    # TODO: Do not hardcode max_mods
    # TODO: Do not include fixed modifications in max_mods
    mods_sidechain, mods_nterm = get_modifications_by_target(params["modifications"])
    mod_versions = get_modification_versions(
        row["peptide"], params["modifications"], mods_sidechain, mods_nterm, max_mods=3
    )

    df_out = pd.DataFrame(columns=row.index)
    df_out["modifications"] = ["-" if not mods else mods for mods in mod_versions]
    df_out["spec_id"] = [
        "{}_{:03d}".format(row["spec_id"], i) for i in range(len(mod_versions))
    ]
    df_out["charge"] = row["charge"]
    df_out["peptide"] = row["peptide"]
    if "protein_list" in row.index:
        df_out["protein_list"] = str(row["protein_list"])
    return df_out


def add_charges(df_in):
    params = get_params()
    df_out = pd.DataFrame(columns=df_in.columns)
    for charge in params["charges"]:
        tmp = df_in.copy()
        tmp["spec_id"] = tmp["spec_id"] + "_{}".format(charge)
        tmp["charge"] = charge
        df_out = pd.concat([df_out, tmp], axis=0, ignore_index=True)
    df_out.sort_values(["spec_id", "charge"], inplace=True)
    df_out.reset_index(drop=True, inplace=True)
    return df_out


def create_decoy_peprec(
    peprec,
    spec_id_prefix="decoy_",
    keep_cterm_aa=True,
    remove_redundancy=True,
    move_mods=True,
):
    """
    Create decoy peptides by reversing the sequences in a PEPREC DataFrame.

    Keyword arguments:
    spec_id_prefix -- string to prefix the decoy spec_ids (default: 'decoy_')
    keep_cterm_aa -- True if the last amino acid should stay in place (for example to keep tryptic properties) (default: True)
    remove_redundancy -- True if reversed peptides that are also found in the set of normal peptide should be removed (default: True)
    move_mods -- True to move modifications according to reversed sequence (default: True)

    Known issues:
    - C-terminal modifications (with position `-1`) are sorted to the front (eg: `-1|Cterm|0|Nterm|2|NormalPTM`).
    """

    def move_mods(row):
        mods = row["modifications"]
        if type(mods) == str:
            if not mods == "-":
                mods = mods.split("|")
                mods = sorted(
                    zip(
                        [
                            int(p)
                            if (p == "-1" or p == "0")
                            else len(row["peptide"]) - int(p)
                            for p in mods[::2]
                        ],
                        mods[1::2],
                    )
                )
                mods = "|".join(["|".join([str(x) for x in mod]) for mod in mods])
                row["modifications"] = mods
        return row

    peprec_decoy = peprec.copy()
    peprec_decoy["spec_id"] = spec_id_prefix + peprec_decoy["spec_id"].astype(str)

    if keep_cterm_aa:
        peprec_decoy["peptide"] = peprec_decoy["peptide"].apply(
            lambda pep: pep[-2::-1] + pep[-1]
        )
    else:
        peprec_decoy["peptide"] = peprec_decoy["peptide"].apply(lambda pep: pep[-1::-1])

    if remove_redundancy:
        peprec_decoy = peprec_decoy[~peprec_decoy["peptide"].isin(peprec["peptide"])]

    if "protein_list" in peprec_decoy.columns:
        peprec_decoy["protein_list"] = "decoy"

    if move_mods:
        peprec_decoy = peprec_decoy.apply(move_mods, axis=1)

    return peprec_decoy


def remove_from_peprec_filter(peprec_pred, peprec_filter):
    peprec_pred_comb = (
        peprec_pred["modifications"]
        + peprec_pred["peptide"]
        + peprec_pred["charge"].astype(str)
    )
    peprec_filter_comb = (
        peprec_filter["modifications"]
        + peprec_filter["peptide"]
        + peprec_filter["charge"].astype(str)
    )
    return peprec_pred[~peprec_pred_comb.isin(peprec_filter_comb)].copy()


def run_batches(peprec, decoy=False):
    params = get_params()
    if decoy:
        params["output_filename"] += "_decoy"

    ms2pip_params = {
        "ms2pip": {
            "model": params["ms2pip_model"],
            "frag_error": 0.02,
            # Modify fasta2speclib modifications dict to MS2PIP params PTMs entry
            "ptm": [
                "{},{},opt,{}".format(
                    mods["name"], mods["mass_shift"], mods["amino_acid"]
                )
                if not mods["n_term"]
                else "{},{},opt,N-term".format(mods["name"], mods["mass_shift"])
                for mods in params["modifications"]
            ],
            "sptm": [],
            "gptm": [],
        }
    }

    # If add_retention_time, initiate DeepLC
    if params["add_retention_time"]:
        logging.debug("Initializing DeepLC predictor")
        if "deeplc" not in params or not params["deeplc"]:
            params["deeplc"] = {"calibration_file": None}
        if not "n_jobs" in params["deeplc"]:
            params["deeplc"]["n_jobs"] = params["num_cpu"]
        rt_predictor = RetentionTime(config=params)

    # Split up into batches to save memory:
    b_size = params["batch_size"]
    b_count = 0
    num_b_counts = ceil(len(peprec) / b_size)
    for i in range(0, len(peprec), b_size):
        if i + b_size < len(peprec):
            peprec_batch = peprec[i : i + b_size]
        else:
            peprec_batch = peprec[i:]
        b_count += 1
        logging.info(
            "Predicting batch %d of %d, containing %d unmodified peptides",
            b_count,
            num_b_counts,
            len(peprec_batch),
        )

        logging.debug("Adding all modification combinations")
        peprec_mods = pd.DataFrame(columns=peprec_batch.columns)
        with multiprocessing.Pool(params["num_cpu"]) as p:
            peprec_mods = pd.concat(
                [peprec_mods] + p.map(add_mods, peprec_batch.iterrows()),
                ignore_index=True,
            )
        peprec_batch = peprec_mods

        if params["add_retention_time"]:
            logging.info("Adding DeepLC predicted retention times")
            rt_predictor.add_rt_predictions(peprec_batch)
        elif type(params["elude_model_file"]) == str:
            logging.debug("Adding ELUDE predicted retention times")
            peprec_batch["rt"] = get_elude_predictions(
                peprec_batch,
                params["elude_model_file"],
                unimod_mapping={
                    mod["name"]: mod["unimod_accession"]
                    for mod in params["modifications"]
                },
            )

        if type(params["rt_predictions_file"]) == str:
            logging.info("Adding RT predictions from file")
            rt_df = pd.read_csv(params["rt_predictions_file"])
            for col in ["peptide", "modifications", "rt"]:
                assert col in rt_df.columns, (
                    "RT file should contain a `%s` column" % col
                )
            peprec_batch = peprec_batch.merge(
                rt_df, on=["peptide", "modifications"], how="left"
            )
            assert (
                not peprec_batch["rt"].isna().any()
            ), "Not all required peptide-modification combinations could be found in RT file"

        logging.debug("Adding charge states %s", str(params["charges"]))
        peprec_batch = add_charges(peprec_batch)

        if type(params["peprec_filter"]) == str:
            logging.debug("Removing peptides present in peprec filter")
            peprec_filter = pd.read_csv(params["peprec_filter"], sep=" ")
            peprec_batch = remove_from_peprec_filter(peprec_batch, peprec_filter)

        if params["save_peprec"]:
            peprec_batch.to_csv(params["output_filename"] + "_" + str(b_count) + ".csv")

        logging.info("Running MS2PIP for %d peptides", len(peprec_batch))
        ms2pip = MS2PIP(
            peprec_batch,
            num_cpu=params["num_cpu"],
            output_filename=params["output_filename"],
            params=ms2pip_params,
            return_results=True,
        )
        all_preds = ms2pip.run()

        if b_count == 1:
            write_mode = "w"
            append = False
        else:
            write_mode = "a"
            append = True

        if "hdf" in params["output_filetype"]:
            logging.info(
                "Writing predictions to %s_predictions.hdf", params["output_filename"]
            )
            all_preds.astype(str).to_hdf(
                "{}_predictions.hdf".format(params["output_filename"]),
                key="table",
                format="table",
                complevel=3,
                complib="zlib",
                mode=write_mode,
                append=append,
                min_itemsize=50,
            )

        spec_out = spectrum_output.SpectrumOutput(
            all_preds,
            peprec_batch,
            ms2pip_params["ms2pip"],
            output_filename="{}".format(params["output_filename"]),
            write_mode=write_mode,
        )

        if "msp" in params["output_filetype"]:
            logging.info("Writing MSP file")
            spec_out.write_msp()

        if "mgf" in params["output_filetype"]:
            logging.info("Writing MGF file")
            spec_out.write_mgf()

        if "bibliospec" in params["output_filetype"]:
            logging.info("Writing BiblioSpec SSL and MS2 files")
            spec_out.write_bibliospec()

        if "spectronaut" in params["output_filetype"]:
            logging.info("Writing Spectronaut CSV file")
            spec_out.write_spectronaut()

        if "dlib" in params["output_filetype"]:
            logging.info("Writing DLIB SQLite file")
            spec_out.write_dlib()

        del all_preds
        del peprec_batch


def main():
    params = get_params()
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        level=params["log_level"],
    )
    peprec = pd.DataFrame(columns=["spec_id", "peptide", "modifications", "charge"])

    logging.info("Cleaving proteins, adding peptides to peprec")
    with multiprocessing.Pool(params["num_cpu"]) as p:
        peprec = pd.concat(
            [peprec]
            + p.map(prot_to_peprec, SeqIO.parse(params["fasta_filename"], "fasta")),
            ignore_index=True,
        )

    logging.info("Removing peptide redundancy, adding protein list to peptides")
    peprec = get_protein_list(peprec)

    peprec_nonmod = peprec.copy()

    save_peprec = False
    if save_peprec:
        logging.info(
            "Saving non-expanded PEPREC to %s.peprec.hdf", params["output_filename"]
        )
        peprec_nonmod["protein_list"] = [
            "/".join(prot) for prot in peprec_nonmod["protein_list"]
        ]
        peprec_nonmod.astype(str).to_hdf(
            "{}_nonexpanded.peprec.hdf".format(params["output_filename"]),
            key="table",
            format="table",
            complevel=3,
            complib="zlib",
            mode="w",
        )

    if not params["decoy"]:
        del peprec_nonmod

    run_batches(peprec, decoy=False)

    if params["decoy"]:
        logging.info("Reversing sequences for decoy peptides")
        peprec_decoy = create_decoy_peprec(peprec_nonmod, move_mods=False)
        del peprec_nonmod

        logging.info("Predicting spectra for decoy peptides")
        run_batches(peprec_decoy, decoy=True)

    logging.info("fasta2speclib is ready!")


if __name__ == "__main__":
    main()



#cython: language_level=3
import sys
import numpy as np
cimport numpy as np


NUM_ION_TYPES_MAPPING = {'general': 2, 'etd': 4, 'ch2': 4, 'all': 18}


cdef extern from "ms2pip_peaks_c.c":

	ctypedef struct annotations:
		float* peaks
		float* msms

	void init_ms2pip(char* amino_masses_fname, char* modifications_fname, char* modifications_fname_sptm)

	unsigned int* get_v_ms2pip(int peplen, unsigned short* peptide, unsigned short* modpeptide, int charge)
	unsigned int* get_v_ms2pip_ce(int peplen, unsigned short* peptide, unsigned short* modpeptide, int charge, int ce)
	unsigned int* get_v_ms2pip_old(int peplen, unsigned short* peptide, unsigned short* modpeptide, int charge)
	unsigned int* get_v_ms2pip_catboost(int peplen, unsigned short* peptide, unsigned short* modpeptide, int charge)

	float* get_p_ms2pip(int peplen, unsigned short* peptide, unsigned short* modpeptide, int charge, int model_id, int ce)

	float* get_mz_ms2pip_general(int peplen, unsigned short* modpeptide)
	float* get_mz_ms2pip_etd(int peplen, unsigned short* modpeptide)
	float* get_mz_ms2pip_ch2(int peplen, unsigned short* modpeptide)

	annotations get_t_ms2pip_all(int peplen, unsigned short* modpeptide, int numpeaks, float* msms, float* peaks, float tolmz)
	float* get_t_ms2pip_general(int peplen, unsigned short* modpeptide, int numpeaks, float* msms, float* peaks, float tolmz)
	float* get_t_ms2pip_etd(int peplen, unsigned short* modpeptide, int numpeaks, float* msms, float* peaks, float tolmz)
	float* get_t_ms2pip_ch2(int peplen, unsigned short* modpeptide, int numpeaks, float* msms, float* peaks, float tolmz)


def ms2pip_init(amino_masses_fname, modifications_fname, modifications_fname_sptm):
	if not isinstance(amino_masses_fname, bytearray):
		amino_masses_fname = bytearray(amino_masses_fname.encode())
	if not isinstance(modifications_fname, bytearray):
		modifications_fname = bytearray(modifications_fname.encode())
	if not isinstance(modifications_fname_sptm, bytearray):
		modifications_fname_sptm = bytearray(modifications_fname_sptm.encode())
	init_ms2pip(amino_masses_fname, modifications_fname, modifications_fname_sptm)


def get_vector(np.ndarray[unsigned short, ndim=1, mode="c"] peptide,
			   np.ndarray[unsigned short, ndim=1, mode="c"] modpeptide,
			   charge):

	cdef unsigned int* results = get_v_ms2pip(len(peptide)-2, &peptide[0], &modpeptide[0], charge)

	r = []
	offset = 0
	fnum = int(results[0] / (len(peptide) - 3))
	for i in range(len(peptide) - 3):
		v = []
		for j in range(fnum):
			v.append(results[j + 1 + offset])
		offset += fnum
		r.append(np.array(v, dtype=np.uint16))

	return r


def get_vector_ce(np.ndarray[unsigned short, ndim=1, mode="c"] peptide,
			   np.ndarray[unsigned short, ndim=1, mode="c"] modpeptide,
			   charge, ce):

	cdef unsigned int* results = get_v_ms2pip_ce(len(peptide)-2, &peptide[0], &modpeptide[0], charge, ce)

	r = []
	offset = 0
	fnum = int(results[0] / (len(peptide) - 3))
	for i in range(len(peptide) - 3):
		v = []
		for j in range(fnum):
			v.append(results[j + 1 + offset])
		offset += fnum
		r.append(np.array(v, dtype=np.uint16))

	return r


def get_vector_catboost(np.ndarray[unsigned short, ndim=1, mode="c"] peptide,
						np.ndarray[unsigned short, ndim=1, mode="c"] modpeptide,
						charge):

	cdef unsigned int* results = get_v_ms2pip_catboost(len(peptide)-2, &peptide[0], &modpeptide[0], charge)

	r = []
	offset = 0
	fnum = int(results[0] / (len(peptide) - 3))
	for i in range(len(peptide) - 3):
		v = []
		for j in range(fnum):
			v.append(results[j + 1 + offset])
		offset += fnum
		r.append(np.array(v, dtype=np.uint16))

	return r


def get_predictions(np.ndarray[unsigned short, ndim=1, mode="c"] peptide,
					np.ndarray[unsigned short, ndim=1, mode="c"] modpeptide,
					charge, model_id, peaks_version, ce):
	cdef float* results = get_p_ms2pip(len(peptide)-2, &peptide[0], &modpeptide[0], charge, model_id, ce)
	if results is NULL:
		raise NotImplementedError(model_id)
	result_parsed = []
	for i in range(NUM_ION_TYPES_MAPPING[peaks_version]):
		tmp = []
		for j in range(len(modpeptide)-3):
			tmp.append(results[(len(modpeptide)-3) * i + j])
		result_parsed.append(tmp)
	return result_parsed


def get_targets(*args):
	if args[4] == 'general':
		result = get_targets_general(*args)
	if args[4] == 'etd':
		result = get_targets_etd(*args)
	if args[4] == 'ch2':
		result = get_targets_ch2(*args)
	if args[4] == 'all':
		result = get_targets_all(*args)
	return result


def get_targets_all(np.ndarray[unsigned short, ndim=1, mode="c"] modpeptide,
						np.ndarray[float, ndim=1, mode="c"] msms,
						np.ndarray[float, ndim=1, mode="c"] peaks,
						fragerror, peaks_version):
	cdef annotations results
	results = get_t_ms2pip_all(len(modpeptide)-2, &modpeptide[0], len(peaks), &msms[0], &peaks[0], fragerror)
	result_peaks = []
	for i in range(NUM_ION_TYPES_MAPPING[peaks_version]*(len(modpeptide)-3)):
		result_peaks.append(results.peaks[i])
	result_mzs = []
	for i in range(NUM_ION_TYPES_MAPPING[peaks_version]*(len(modpeptide)-3)):
		result_mzs.append(results.msms[i])
	#print(result_parsed)
	#print()
	return (result_mzs,result_peaks)

def get_targets_general(np.ndarray[unsigned short, ndim=1, mode="c"] modpeptide,
						np.ndarray[float, ndim=1, mode="c"] msms,
						np.ndarray[float, ndim=1, mode="c"] peaks,
						fragerror, peaks_version):
	cdef float* results = get_t_ms2pip_general(len(modpeptide)-2, &modpeptide[0], len(peaks), &msms[0], &peaks[0], fragerror)
	result_parsed = []
	for i in range(NUM_ION_TYPES_MAPPING[peaks_version]): #SD: HAd to change this
		tmp = []
		for j in range(len(modpeptide)-3):
			tmp.append(results[(len(modpeptide)-3) * i + j])
		result_parsed.append(tmp)
	return result_parsed

def get_targets_etd(np.ndarray[unsigned short, ndim=1, mode="c"] modpeptide,
						np.ndarray[float, ndim=1, mode="c"] msms,
						np.ndarray[float, ndim=1, mode="c"] peaks,
						fragerror, peaks_version):
	cdef float* results = get_t_ms2pip_etd(len(modpeptide)-2, &modpeptide[0], len(peaks), &msms[0], &peaks[0], fragerror)
	result_parsed = []
	for i in range(NUM_ION_TYPES_MAPPING[peaks_version]):
		tmp = []
		for j in range(len(modpeptide)-3):
			tmp.append(results[(len(modpeptide)-3) * i + j])
		result_parsed.append(tmp)
	return result_parsed


def get_targets_ch2(np.ndarray[unsigned short, ndim=1, mode="c"] modpeptide,
						np.ndarray[float, ndim=1, mode="c"] msms,
						np.ndarray[float, ndim=1, mode="c"] peaks,
						fragerror, peaks_version):
	cdef float* results = get_t_ms2pip_ch2(len(modpeptide)-2, &modpeptide[0], len(peaks), &msms[0], &peaks[0], fragerror)
	result_parsed = []
	for i in range(NUM_ION_TYPES_MAPPING[peaks_version]):
		tmp = []
		for j in range(len(modpeptide)-3):
			tmp.append(results[(len(modpeptide)-3) * i + j])
		result_parsed.append(tmp)
	return result_parsed


def get_mzs(*args):
	if args[1] == 'general':
		result = get_mzs_general(*args)
	if args[1] == 'etd':
		result = get_mzs_etd(*args)
	if args[1] == 'ch2':
		result = get_mzs_ch2(*args)
	return result


def get_mzs_general(np.ndarray[unsigned short, ndim=1, mode="c"] modpeptide,
					peaks_version):
	cdef float* results = get_mz_ms2pip_general(len(modpeptide)-2, &modpeptide[0])
	result_parsed = []
	for i in range(NUM_ION_TYPES_MAPPING[peaks_version]):
		tmp = []
		for j in range(len(modpeptide)-3):
			tmp.append(results[(len(modpeptide)-3) * i + j])
		result_parsed.append(tmp)
	return result_parsed


def get_mzs_etd(np.ndarray[unsigned short, ndim=1, mode="c"] modpeptide,
				peaks_version):
	cdef float* results = get_mz_ms2pip_etd(len(modpeptide)-2, &modpeptide[0])
	result_parsed = []
	for i in range(NUM_ION_TYPES_MAPPING[peaks_version]):
		tmp = []
		for j in range(len(modpeptide)-3):
			tmp.append(results[(len(modpeptide)-3) * i + j])
		result_parsed.append(tmp)
	return result_parsed


def get_mzs_ch2(np.ndarray[unsigned short, ndim=1, mode="c"] modpeptide,
				peaks_version):
	cdef float* results = get_mz_ms2pip_ch2(len(modpeptide)-2, &modpeptide[0])
	result_parsed = []
	for i in range(NUM_ION_TYPES_MAPPING[peaks_version]):
		tmp = []
		for j in range(len(modpeptide)-3):
			tmp.append(results[(len(modpeptide)-3) * i + j])
		result_parsed.append(tmp)
	return result_parsed



"""
MS2PIP / PEPREC modification handling
"""

import itertools
import tempfile

from pyteomics import mass

from ms2pip.exceptions import InvalidAminoAcidError

AMINO_ACIDS = [
    "A",
    "C",
    "D",
    "E",
    "F",
    "G",
    "H",
    "I",
    "K",
    "M",
    "N",
    "P",
    "Q",
    "R",
    "S",
    "T",
    "V",
    "W",
    "Y",
]

AMINO_ACID_MASSES = [
    71.037114,
    103.00919,
    115.026943,
    129.042593,
    147.068414,
    57.021464,
    137.058912,
    113.084064,
    128.094963,
    131.040485,
    114.042927,
    97.052764,
    128.058578,
    156.101111,
    87.032028,
    101.047679,
    99.068414,
    186.079313,
    163.063329,
    # 147.0354  # iTRAQ fixed N-term modification (gets written to amino acid masses file)
]

AMINO_ACID_IDS = {a: i for i, a in enumerate(AMINO_ACIDS)}

PROTON_MASS = 1.007825032070059


class Modifications:
    def __init__(self):
        """
        MS2PIP / PEPREC modification handling
        """
        self.modifications = {}
        self._mass_shifts = None
        self._ptm_ids = None
        self._next_mod_id = 38  # Omega compatibility (mutations)

    def add_from_ms2pip_modstrings(self, modstrings, mod_type="ptm"):
        """
        Add modifications from MS2PIP modstring list

        Parameters
        ----------
        modstrings: list(str)
            List of MS2PIP modstrings

        Example
        -------
        >>> ms2pip_ptms = [
        ...     "Oxidation,15.994915,opt,M",
        ...     "Acetyl,42.010565,opt,N-term",
        ... ]
        ... mods = Modifications()
        ... mods.add_from_ms2pip_modstrings(ms2pip_ptms)
        """

        if mod_type not in self.modifications:
            self.modifications[mod_type] = {}

        # NOTE: opt is ignored
        for mod in modstrings:
            mod_name, mass_shift, opt, amino_acid = mod.split(",")

            if amino_acid == "N-term":
                amino_acid_id = -1
            elif amino_acid == "C-term":
                amino_acid_id = -2
            elif amino_acid == "L":
                amino_acid_id = AMINO_ACID_IDS["I"]
            elif amino_acid in AMINO_ACID_IDS:
                amino_acid_id = AMINO_ACID_IDS[amino_acid]
            else:
                raise InvalidAminoAcidError(amino_acid)

            self.modifications[mod_type][mod_name] = {
                "mass_shift": float(mass_shift),
                "amino_acid": amino_acid,
                "amino_acid_id": amino_acid_id,
                "mod_id": self._next_mod_id,
            }
            self._next_mod_id += 1

        self._mass_shifts = None

    @property
    def _all_modifications(self):
        return itertools.chain.from_iterable(
            (mods.items() for mods in self.modifications.values())
        )

    @property
    def mass_shifts(self):
        """
        Return modification name -> mass shift mapping.
        """
        if not self._mass_shifts:
            self._mass_shifts = {
                name: mod["mass_shift"] for name, mod in self._all_modifications
            }
        return self._mass_shifts

    @property
    def ptm_ids(self):
        """
        Return modification name -> modification id mapping.
        """
        if not self._ptm_ids:
            self._ptm_ids = {
                name: mod["mod_id"] for name, mod in self._all_modifications
            }
        return self._ptm_ids

    def write_modifications_file(self, mod_type="ptm"):
        mod_file = tempfile.NamedTemporaryFile(delete=False, mode="w", newline="\n")
        mod_file.write("{}\n".format(len(self.modifications[mod_type])))
        for name, mod in self.modifications[mod_type].items():
            mod_file.write(
                "{},1,{},{}\n".format(
                    mod["mass_shift"], mod["amino_acid_id"], mod["mod_id"]
                )
            )
        mod_file.close()
        return mod_file.name

    def calc_precursor_mz(self, peptide, modifications, charge):
        """
        Calculate precursor mass and mz for given peptide and modification list,
        taking the modifications into account.

        Note: This method does not use the build-in Pyteomics modification handling, as
        that would require a known atomic composition of the modification.

        Parameters
        ----------
        peptide: str
            stripped peptide sequence

        modifications: str
            MS2PIP-style formatted modifications list (e.g. `0|Acetyl|2|Oxidation`)

        charge: int
            precursor charge

        Returns
        -------
        prec_mass, prec_mz: tuple(float, float)
        """

        charge = int(charge)
        unmodified_mass = mass.fast_mass(peptide)
        mods_massses = sum(
            [self.mass_shifts[mod] for mod in modifications.split("|")[1::2]]
        )
        prec_mass = unmodified_mass + mods_massses
        prec_mz = (prec_mass + charge * PROTON_MASS) / charge
        return prec_mass, prec_mz


def write_amino_acid_masses():
    # Includes fixed/variable information for Omega compatibility
    amino_file = tempfile.NamedTemporaryFile(delete=False, mode="w", newline="\n")
    for m in AMINO_ACID_MASSES:
        amino_file.write("{}\n".format(m))
    amino_file.write("0\n")
    amino_file.close()
    return amino_file.name



"""Get predictions directly from XGBoost model, within ms2pip framework."""

import hashlib
import logging
import os
import urllib.request
from itertools import islice

import numpy as np
import xgboost as xgb

from ms2pip.exceptions import InvalidXGBoostModelError

logger = logging.getLogger(__name__)


def get_predictions_xgb(features, num_ions, model_params, model_dir, num_cpu=1):
    """
    Get predictions starting from feature vectors in DMatrix object.

    Parameters
    ----------
    features: xgb.DMatrix
        Feature vectors in DMatrix object
    num_ions: list[int]
        List with number of ions (per series) for each peptide, i.e. peptide length - 1
    model_params: dict
        Model configuration as defined in ms2pip.ms2pipC.MODELS.
    model_dir: str
        Directory where model files are stored.
    num_cpu: int
        Number of CPUs to use in multiprocessing

    """
    # Init models
    xgboost_models = initialize_xgb_models(
        model_params["xgboost_model_files"],
        model_dir,
        num_cpu,
    )

    logger.debug("Predicting intensities from XGBoost model files...")
    preds_list = []
    for ion_type, xgb_model in xgboost_models.items():
        # Get predictions from XGBoost model
        preds = xgb_model.predict(features)
        xgb_model.__del__()

        # Reshape into arrays for each peptide
        preds = _split_list_by_lengths(preds, num_ions)
        if ion_type in ["x", "y", "y2", "z"]:
            preds = [np.array(x[::-1], dtype=np.float32) for x in preds]
        elif ion_type in ["a", "b", "b2", "c"]:
            preds = [np.array(x, dtype=np.float32) for x in preds]
        else:
            raise ValueError(f"Unsupported ion_type: {ion_type}")
        preds_list.append(preds)

    predictions = [list(t) for t in zip(*preds_list)]
    return predictions


def _split_list_by_lengths(list_in, lengths):
    list_in = iter(list_in)
    return [list(islice(list_in, elem)) for elem in lengths]


def check_model_presence(model, model_hash, model_dir):
    """Check whether XGBoost model file is downloaded."""
    filename = os.path.join(model_dir, model)
    if not os.path.isfile(filename):
        return False
    return check_model_integrity(filename, model_hash)


def download_model(model, model_hash, model_dir):
    """Download the xgboost model from the Genesis server."""
    if not os.path.isdir(model_dir):
        os.mkdir(model_dir)
    filename = os.path.join(model_dir, model)

    logger.info(f"Downloading {model} to {filename}...")
    urllib.request.urlretrieve(
        os.path.join("http://genesis.ugent.be/uvpublicdata/ms2pip/", model), filename
    )
    if not check_model_integrity(filename, model_hash):
        raise InvalidXGBoostModelError()


def check_model_integrity(filename, model_hash):
    """Check that models are correctly downloaded."""
    sha1_hash = hashlib.sha1()
    with open(filename, "rb") as model_file:
        while True:
            chunk = model_file.read(16 * 1024)
            if not chunk:
                break
            sha1_hash.update(chunk)
    if sha1_hash.hexdigest() == model_hash:
        return True
    else:
        logger.warn("Model hash not recognized.")
        return False


def validate_requested_xgb_model(xgboost_model_files, xgboost_model_hashes, model_dir):
    """Validate requested XGBoost models, and download if necessary"""
    for _, model_file in xgboost_model_files.items():
        if not check_model_presence(
            model_file, xgboost_model_hashes[model_file], model_dir
        ):
            download_model(model_file, xgboost_model_hashes[model_file], model_dir)


def initialize_xgb_models(xgboost_model_files, model_dir, nthread) -> dict:
    """Initialize xgboost models and return them in a dict with ion types as keys."""
    xgb.set_config(verbosity=0)
    xgboost_models = {}
    for ion_type in xgboost_model_files.keys():
        model_file = os.path.join(model_dir, xgboost_model_files[ion_type])
        logger.debug(f"Initializing model from file: `{model_file}`")
        xgb_model = xgb.Booster({"nthread": nthread})
        xgb_model.load_model(model_file)
        xgboost_models[ion_type] = xgb_model
    return xgboost_models



import os

import pandas as pd

# from deeplc import DeepLC

# Reduce Tensorflow logging
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"

class RetentionTime:
    def __init__(self, predictor="deeplc", config=None):
        """
        Initialize peptide retention time predictor

        Parameters
        ----------
        predictor: str, optional
            Retention time predictor to employ. Currently only 'deeplc' is supported.
        config: dict, optional
            Dictionary with configuration. Requires 'deeplc' top-level key for
            DeepLC predictions.
        """
        self.predictor = predictor
        self.deeplc_predictor = None

        if not config:
            self.config = dict()
        else:
            self.config = config

    def _get_irt_peptides(self):
        """
        Return DeepLC DataFrame with iRT peptides
        """
        irt_peptides = {
            "LGGNEQVTR": -24.92,
            "GAGSSEPVTGLDAK": 0.00,
            "VEATFGVDESNAK": 12.39,
            "YILAGVENSK": 19.79,
            "TPVISGGPYEYR": 28.71,
            "TPVITGAPYEYR": 33.38,
            "DGLDAASYYAPVR": 42.26,
            "ADVTPADFSEWSK": 54.62,
            "GTFIIDPGGVIR": 70.52,
            "GTFIIDPAAVIR": 87.23,
            "LFLQFGAQGSPFLK": 100.00,
        }

        irt_df = pd.DataFrame.from_dict(irt_peptides, orient="index")
        irt_df = irt_df.reset_index()
        irt_df.columns = ["seq", "tr"]
        irt_df["modifications"] = ""

        return irt_df

    def _init_deeplc(self):
        """
        Initialize DeepLC: import, configurate and calibrate
        """
        # Only import if DeepLC will be used, otherwise lot's of extra heavy
        # dependencies (e.g. Tensorflow) are imported as well
        from deeplc import DeepLC

        if "deeplc" in self.config:
            deeplc_params = self.config["deeplc"]
            calibration_file = deeplc_params["calibration_file"]
            del deeplc_params["calibration_file"]
        else:
            deeplc_params = {"verbose": False}
            calibration_file = None

        # TODO: Remove when fixed upstream in DeepLC
        if not calibration_file:
            deeplc_params["split_cal"] = 9

        self.deeplc_predictor = DeepLC(**deeplc_params)

        if calibration_file:
            cal_df = pd.read_csv(calibration_file, sep=",")
        else:
            cal_df = self._get_irt_peptides()

        self.deeplc_predictor.calibrate_preds(seq_df=cal_df)

    def _prepare_deeplc_peptide_df(self):
        """
        Prepare DeepLC peptide DataFrame
        """
        column_map = {"peptide": "seq", "modifications": "modifications"}
        self.deeplc_pep_df = self.peprec[column_map.keys()].copy()
        self.deeplc_pep_df.rename(columns=column_map, inplace=True)

    def _run_deeplc(self):
        """
        Run DeepLC
        """
        self.deeplc_preds = self.deeplc_predictor.make_preds(
            seq_df=self.deeplc_pep_df.fillna("")
        )

    def _parse_deeplc_preds(self):
        """
        Add DeepLC predictions to peprec DataFrame
        """
        self.peprec["rt"] = self.deeplc_preds

    def _predict_deeplc(self):
        """
        Predict retention times using DeepLC
        """
        if not self.deeplc_predictor:
            self._init_deeplc()
        self._prepare_deeplc_peptide_df()
        self._run_deeplc()
        self._parse_deeplc_preds()

    def add_rt_predictions(self, peprec):
        """
        Run RT predictor and add predictions to peprec DataFrame.

        peprec: pandas.DataFrame
            MS2PIP-style peprec DataFrame with peptides for which to predict retention
            times
        """
        self.peprec = peprec

        if self.predictor == "deeplc":
            self._predict_deeplc()
        else:
            raise NotImplementedError(self.predictor)



import numpy as np


def calc_correlations(df):
    correlations = (
        df.groupby(["spec_id", "ion"])[["target", "prediction"]]
        .corr()
        .iloc[::2]["prediction"]
    )
    correlations.index = correlations.index.droplevel(2)
    correlations = correlations.to_frame().reset_index()
    correlations.columns = ["spec_id", "ion", "pearsonr"]
    return correlations


def ms2pip_pearson(true, pred):
    """
    Return pearson of tic-normalized, log-transformed intensities,
    the MS2PIP way.
    """
    tic_norm = lambda x: x / np.sum(x)
    log_transform = lambda x: np.log2(x + 0.001)
    corr = np.corrcoef(log_transform(tic_norm(true)), log_transform(tic_norm(pred)))[0][
        1
    ]
    return corr


def spectral_angle(true, pred, epsilon=1e-7):
    """
    Return square root normalized spectral angle.
    See https://doi.org/10.1074/mcp.O113.036475
    """
    pred_norm = pred / max(np.linalg.norm(pred), epsilon)
    true_norm = true / max(np.linalg.norm(true), epsilon)
    spectral_angle = 1 - (2 * np.arccos(np.dot(pred_norm, true_norm)) / np.pi)
    return spectral_angle



"""
ELUDE RT prediction integration for MS2PIP

ELUDE needs to be installed and callable with os.system(elude). Tested with
ELUDE v3.2.

Can be run as a script or by calling the `get_elude_predictions` function.
Both the script and the function take an MS2PIP PEPREC file and an ELUDE model
file.

If the given PEPREC contains PTMs, these need to be configured in the (for now)
hardcoded dictionary `unimod_mapping`. This is a dictionary where the keys
represent the name of the PTM in the PEPREC file and the values contain the
corresponding UniMod IDs.
"""


__author__ = "Ralf Gabriels"
__credits__ = ["Ralf Gabriels", "Sven Degroeve", "Lennart Martens"]
__license__ = "Apache License, Version 2.0"
__version__ = "0.1"
__email__ = "Ralf.Gabriels@ugent.be"


# Standard library
import os
import argparse

# Third party packages
import pandas as pd


def argument_parser():
	parser = argparse.ArgumentParser(description='Add ELUDE retention time predictions to existing PEPREC file.')
	parser.add_argument("peprec_file", metavar="<peprec file>",
						help="MS2PIP PEPREC file to add RT predictions to.")
	parser.add_argument('elude_model_file', metavar="<elude model file>",
						help='ELUDE model file.')
	args = parser.parse_args()
	return args


def elude_insert_mods(row, peptide_column='peptide', mods_column='modifications',
					  unimod_mapping=None):
	"""
	Insert PEPREC modifications into peptide sequence for ELUDE.

	Accepts normal, N-terminal and C-terminal modifications.

	Positional arguments:
	row -- pandas.DataFrame row, for use with .apply(elude_insert_mods, axis=1)

	Keyword arguments:
	peptide_column -- Column name of column with peptide sequences
	mods_column -- Column name of column with MS2PIP PEPREC encoded
	modifications
	unimod_mapping -- Dictionary that maps the MS2PIP modification names to
	UniMod accessions
	"""

	if not unimod_mapping:
		unimod_mapping = {'Oxidation': 35, 'Carbamidomethyl': 4}

	peptide = row[peptide_column]
	mods = row[mods_column]

	if type(mods) == str:
		if mods == '-':
			peptide_mods = peptide

		else:
			mods = mods.split('|')
			pos, names = [list(tup) for tup in zip(*sorted(zip([int(p) for p in mods[::2]], mods[1::2])))]

			# Add 0 to pos, if no N-term PTMs present
			if pos[0] != 0:
				pos.insert(0, 0)
				names.insert(0, '')

			# Replace "-1" index from C-term PTMs to index of last aa
			if pos[-1] == -1:
				pos[-1] = len(peptide)

			pep_split = [peptide[i:j] for i, j in zip(list(pos), list(pos)[1:] + [None])]
			peptide_mods = ''.join([''.join(tup) for tup in zip(['[unimod:{}]'.format(unimod_mapping[n])
									if n else '' for n in names], pep_split)])

	else:
		peptide_mods = peptide

	return peptide_mods


def get_elude_predictions(peprec, elude_model_file, **kwargs):
	"""
	Return ELUDE retention time predictions for peptide sequences in MS2PIP PEPREC.

	ELUDE needs to be installed and callable with os.system(elude). Tested with ELUDE v3.2

	Positional arguments:
	peprec -- MS2PIP PEPREC in pandas.DataFrame()
	elude_model_file -- filename of ELUDE model to apply

	kwargs -- keyword arguments are passed to elude_insert_mods
	"""

	filename_in = '{}_Test.txt'.format(elude_model_file)
	filename_out = '{}_Preds.txt'.format(elude_model_file)
	filename_model = elude_model_file

	peprec.apply(elude_insert_mods, **kwargs, axis=1)\
		  .to_csv(filename_in, sep=' ', index=False, header=False)
	os.system('elude -l "{}" -e "{}" -o "{}" -p'.format(filename_model, filename_in, filename_out))
	preds = pd.read_csv(filename_out, sep='\t', comment='#')
	os.system('rm {}; rm {}'.format(filename_in, filename_out))

	return preds['Predicted_RT']


def main():
	args = argument_parser()

	# HARD CODED, for now
	unimod_mapping = {
		# PEPREC PTM name: UniMod ID,
		'Oxidation': 35,
		'Carbamidomethyl': 4
	}

	print("Adding ELUDE RT predictions to {}...".format(args.peprec_file))

	peprec = pd.read_csv(
		args.peprec_file, sep=" ", index_col=False,
		dtype={"spec_id": str, "modifications": str}
	)

	# Remove `* 60` to get RTs in seconds
	peprec['rt'] = get_elude_predictions(peprec, args.elude_model_file, unimod_mapping=unimod_mapping) # * 60

	output_filename = '.'.join(args.peprec_file.split('.')[:-1]) + '_with_rt.peprec'
	peprec.to_csv(output_filename, sep=' ', index=False)

	print("Ready!")


if __name__ == '__main__':
	main()



import os

import tomlkit


class UnsupportedConfigFormatError(ValueError):
    pass


class NoFilepathError(Exception):
    pass


class ConfigParser:
    """
    MS2PIP Configuration parser

    Parameters
    ----------
    filepath: str
        Path to config file to load from or write to (optional)
    """

    def __init__(self, filepath=None):
        self.filepath = filepath
        self.config = dict()

        if self.filepath:
            self.load()

    def _set_filepath(self, filepath):
        """
        Set config filepath

        Parameters
        ----------
        filepath: str
            Path to config file to load

        Raises
        ------
        NoFilepathError
            If both filepath and self.filepath are None
        """
        if not filepath:
            if not self.filepath:
                raise NoFilepathError()
        else:
            self.filepath = filepath

    def _load_ms2pip_txt(self):
        params = {}
        params["ptm"] = []
        params["sptm"] = []
        params["gptm"] = []

        with open(self.filepath) as f:
            for line in f:
                line = line.strip()
                if not line or line[0] == "#":
                    continue
                (par, val) = line.split("=")
                if par == "ptm":
                    params["ptm"].append(val)
                elif par == "sptm":
                    params["sptm"].append(val)
                elif par == "gptm":
                    params["gptm"].append(val)
                else:
                    params[par] = val

        if "frag_error" in params:
            params["frag_error"] = float(params["frag_error"])

        self.config["ms2pip"] = params

    def _load_toml(self):
        toml_file = ""
        with open(self.filepath, "rt") as f_in:
            for line in f_in:
                toml_file += line
        self.config = tomlkit.loads(toml_file)

    def _write_toml(self):
        self.filepath = os.path.splitext(self.filepath)[0] + ".toml"
        with open(self.filepath, "wt+") as f_out:
            f_out.write(tomlkit.dumps(self.config))

    def load(self, filepath=None, config_format=None):
        """
        Load configuration file.

        Parameters
        ----------
        filepath: str
            Path to config file to load (optional)
        config_format: str
            Config file format, either `txt` or `toml`. If None, the format will be
            inferred from the filename extension. (optional)
        """
        self._set_filepath(filepath)

        if not config_format:
            config_format = os.path.splitext(self.filepath)[1].lower().lstrip(".")

        if config_format == "toml":
            self._load_toml()
        elif config_format in ("txt", "config", "ms2pip"):
            self._load_ms2pip_txt()
        else:
            raise UnsupportedConfigFormatError(
                "Configuration file should have extension `txt`, `config`, or "
                "`ms2pip` (text-based format) or `toml` (TOML-based format), not "
                f"`{config_format}`",
            )

    def write(self, filepath=None, config_format="toml"):
        """
        Write configuration to file.

        Parameters
        ----------
        filepath: str
            Path where config file will be written (optional)
        config_format: str
            Config file format to write, default: `toml` (optional)
        """
        self._set_filepath(filepath)

        if config_format.lower() == "toml":
            self._write_toml()
        else:
            raise UnsupportedConfigFormatError(config_format)



def get_feature_names():
	"""
	feature names for the fixed peptide length feature vectors
	"""
	aminos = ["A", "C", "D", "E", "F", "G", "H", "I", "K",
			  "M", "N", "P", "Q", "R", "S", "T", "V", "W", "Y"]
	names = []
	for a in aminos:
		names.append("Ib_" + a)
	names.append("sumIbaG")
	names.append("meanIbwikiG")
	names.append("sumIywaG")
	names.append("meanIywikiG")

	names += ["pmz", "peplen", "ionnumber", "ionnumber_rel"]

	for c in ["aG", "wikiG", "mz", "bas", "heli", "hydro", "pI"]:
		names.append("sum_" + c)

	for c in ["mz", "bas", "heli", "hydro", "pI"]:
		names.append("mean_" + c)

	for c in ["max_{}", "min_{}", "max{}_b", "min{}_b", "max{}_y", "min{}_y"]:
		for b in ["bas", "heli", "hydro", "pI"]:
			names.append(c.format(b))

	names.append("mz_ion")
	names.append("mz_ion_other")
	names.append("mean_mz_ion")
	names.append("mean_mz_ion_other")

	for c in ["bas", "heli", "hydro", "pI"]:
		names.append("{}_ion".format(c))
		names.append("{}_ion_other".format(c))
		names.append("{}_ion_minus_ion_other".format(c))
		#names.append("mean_{}_ion".format(c))
		#names.append("mean_{}_ion_other".format(c))

	for c in ["plus_cleave{}", "times_cleave{}", "minus1_cleave{}", "minus2_cleave{}", "bsum{}", "ysum{}"]:
		for b in ["bas", "heli", "hydro", "pI"]:
			names.append(c.format(b))

	for pos in ["0", "1", "-2", "-1"]:
		for c in ["mz", "bas", "heli", "hydro", "pI", "wikiG", "P", "D", "E", "K", "R"]:
			names.append("loc_" + pos + "_" + c)

	for pos in ["i", "i+1"]:
		for c in ["wikiG", "P", "D", "E", "K", "R"]:
			names.append("loc_" + pos + "_" + c)

	for c in ["bas", "heli", "hydro", "pI", "mz"]:
		for pos in ["i", "i-1", "i+1", "i+2"]:
			names.append("loc_" + pos + "_" + c)

	names.append("charge")

	return names


def get_feature_names_catboost():
	num_props = 4
	names = ["amino_first", "amino_last", "amino_lcleave", "amino_rcleave", "peplen", "charge"]
	for t in range(5):
		names.append("charge"+str(t))
	for t in range(num_props):
		names.append("qmin_%i"%t)
		names.append("q1_%i"%t)
		names.append("q2_%i"%t)
		names.append("q3_%i"%t)
		names.append("qmax_%i"%t)
	names.append("len_n")
	names.append("len_c")

	for a in ['A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'M',
			  'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y']:
		names.append("I_n_%s"%a)
		names.append("I_c_%s"%a)

	for t in range(num_props):
		for pos in ["p0", "pend", "pi-1", "pi", "pi+1", "pi+2"]:
			names.append("prop_%i_%s"%(t, pos))
		names.append("sum_%i_n"%t)
		names.append("q0_%i_n"%t)
		names.append("q1_%i_n"%t)
		names.append("q2_%i_n"%t)
		names.append("q3_%i_n"%t)
		names.append("q4_%i_n"%t)
		names.append("sum_%i_c"%t)
		names.append("q0_%i_c"%t)
		names.append("q1_%i_c"%t)
		names.append("q2_%i_c"%t)
		names.append("q3_%i_c"%t)
		names.append("q4_%i_c"%t)

	return names


def get_feature_names_new():
	num_props = 4
	names = ["peplen", "charge"]
	for t in range(5):
		names.append("charge"+str(t))
	for t in range(num_props):
		names.append("qmin_%i"%t)
		names.append("q1_%i"%t)
		names.append("q2_%i"%t)
		names.append("q3_%i"%t)
		names.append("qmax_%i"%t)
	names.append("len_n")
	names.append("len_c")

	for a in ['A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'M',
			  'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y']:
		names.append("I_n_%s"%a)
		names.append("I_c_%s"%a)

	for t in range(num_props):
		for pos in ["p0", "pend", "pi-1", "pi", "pi+1", "pi+2"]:
			names.append("prop_%i_%s"%(t, pos))
		names.append("sum_%i_n"%t)
		names.append("q0_%i_n"%t)
		names.append("q1_%i_n"%t)
		names.append("q2_%i_n"%t)
		names.append("q3_%i_n"%t)
		names.append("q4_%i_n"%t)
		names.append("sum_%i_c"%t)
		names.append("q0_%i_c"%t)
		names.append("q1_%i_c"%t)
		names.append("q2_%i_c"%t)
		names.append("q3_%i_c"%t)
		names.append("q4_%i_c"%t)

	return names


def get_feature_names_small(ionnumber):
	"""
	feature names for the fixed peptide length feature vectors
	"""
	names = []
	names += ["pmz", "peplen"]

	for c in ["bas", "heli", "hydro", "pI"]:
		names.append("sum_" + c)

	for c in ["mz", "bas", "heli", "hydro", "pI"]:
		names.append("mean_" + c)

	names.append("mz_ion")
	names.append("mz_ion_other")
	names.append("mean_mz_ion")
	names.append("mean_mz_ion_other")

	for c in ["bas", "heli", "hydro", "pI"]:
		names.append("{}_ion".format(c))
		names.append("{}_ion_other".format(c))

	names.append("endK")
	names.append("endR")
	names.append("nextP")
	names.append("nextK")
	names.append("nextR")

	for c in ["bas", "heli", "hydro", "pI", "mz"]:
		for pos in ["i", "i-1", "i+1", "i+2"]:
			names.append("loc_" + pos + "_" + c)

	names.append("charge")

	for i in range(ionnumber):
		for c in ["bas", "heli", "hydro", "pI", "mz"]:
			names.append("P_%i_%s"%(i, c))
		names.append("P_%i_P"%i)
		names.append("P_%i_K"%i)
		names.append("P_%i_R"%i)

	return names


def get_feature_names_chem(peplen):
	"""
	feature names for the fixed peptide length feature vectors
	"""

	names = []
	names += ["pmz", "peplen", "ionnumber", "ionnumber_rel", "mean_mz"]

	for c in ["mean_{}", "max_{}", "min_{}", "max{}_b", "min{}_b", "max{}_y", "min{}_y"]:
		for b in ["bas", "heli", "hydro", "pI"]:
			names.append(c.format(b))

	for c in ["mz", "bas", "heli", "hydro", "pI"]:
		names.append("{}_ion".format(c))
		names.append("{}_ion_other".format(c))
		names.append("mean_{}_ion".format(c))
		names.append("mean_{}_ion_other".format(c))

	for c in ["plus_cleave{}", "times_cleave{}", "minus1_cleave{}", "minus2_cleave{}", "bsum{}", "ysum{}"]:
		for b in ["bas", "heli", "hydro", "pI"]:
			names.append(c.format(b))

	for i in range(peplen):
		for c in ["mz", "bas", "heli", "hydro", "pI"]:
			names.append("fix_" + c + "_" + str(i))

	names.append("charge")

	return names



""""Run MS²PIP prediction for single peptide."""

import logging
import os
import re

import click
import matplotlib.pyplot as plt
import numpy as np
import spectrum_utils.plot as sup
import spectrum_utils.spectrum as sus
import xgboost as xgb

from ms2pip.config_parser import ConfigParser
from ms2pip.cython_modules import ms2pip_pyx
from ms2pip.exceptions import InvalidModificationFormattingError, InvalidPeptideError
from ms2pip.ms2pipC import MODELS, apply_mods
from ms2pip.peptides import AMINO_ACID_IDS, Modifications, write_amino_acid_masses
from ms2pip.predict_xgboost import initialize_xgb_models, validate_requested_xgb_model

logger = logging.getLogger("ms2pip")


class SinglePrediction:
    """Run MS²PIP prediction for single peptide."""

    def __init__(self, modification_strings=None, model_dir=None) -> None:
        """
        Run MS²PIP prediction for single peptide.

        Parameters
        ----------
        modification_strings: list-like
            List of MS²PIP configuration-style modification strings, e.g.
            `Carbamidomethyl,57.02146,opt,C` or `Oxidation,15.994915,opt,M`. See MS²PIP
            README.md for more info.
        model_dir : str, optional
            Custom directory for downloaded XGBoost model files. By default,
            `~/.ms2pip` is used.

        Examples
        --------
        >>> from ms2pip.single_prediction import SinglePrediction
        >>> ms2pip_sp = SinglePrediction(
        >>>     modification_strings=[
        >>>         "Carbamidomethyl,57.021464,opt,C"
        >>>     ]
        >>> )
        >>> mz, intensity, annotation = ms2pip_sp.predict(
        >>>     "GSIGECIAEEEEFELDSESNR", "6|Carbamidomethyl", 3
        >>> )

        """
        if not modification_strings:
            modification_strings = []
        self._init_ms2pip(modification_strings)

        if model_dir:
            self.model_dir = model_dir
        else:
            self.model_dir = os.path.join(os.path.expanduser("~"), ".ms2pip")

    def _init_ms2pip(self, modification_strings):
        self.mod_info = Modifications()
        self.mod_info.modifications = {"ptm": {}, "sptm": {}}
        self.mod_info.add_from_ms2pip_modstrings(modification_strings)
        afile = write_amino_acid_masses()
        modfile = self.mod_info.write_modifications_file(mod_type="ptm")
        modfile2 = self.mod_info.write_modifications_file(mod_type="sptm")
        ms2pip_pyx.ms2pip_init(afile, modfile, modfile2)

    def predict(
        self,
        peptide,
        modifications,
        charge,
        model="HCD",
        validate_input=True,
    ):
        """
        Predict single peptide spectrum with MS²PIP.

        Parameters
        ----------
        peptide: string
            Unmodified peptide sequence. Only canonical amino acids are allowed, and
            peptide sequence should be of length [3, 100].
        modifications: string
            MS²PIP style-formatted modification string (e.g. `0|Acetyl|5|Oxidation`).
            See MS²PIP README.md for more info.
        charge: int
            Peptide precursor charge
        model: string (default: "HCD")
            MS²PIP model to use, identical to the option in the MS²PIP configuration
            file.
        validate_input: boolean (default: True)
            Whether to validate input peptide and modifications. Disable for speed-up.

        Returns
        -------
        mz: list[float]
            List with fragment ion m/z values in Da.
        intensity: list[float]
            List with TIC-normalized predicted intensities, order matches `mz`
        annotation: list[str]
            List with fragment ion types and series, order matches `mz`

        """
        peptide = peptide.upper().replace("L", "I")

        if validate_input:
            self._validate_sequence(peptide)
            self._validate_mod_string(modifications)

        peptide = np.array(
            [0] + [AMINO_ACID_IDS[x] for x in peptide] + [0], dtype=np.uint16
        )
        modpeptide = apply_mods(peptide, modifications, self.mod_info.ptm_ids)
        model_id = MODELS[model]["id"]
        peaks_version = MODELS[model]["peaks_version"]
        ce = 30

        mz = np.array(ms2pip_pyx.get_mzs(modpeptide, peaks_version))
        if "xgboost_model_files" in MODELS[model].keys():
            validate_requested_xgb_model(
                MODELS[model]["xgboost_model_files"],
                MODELS[model]["model_hash"],
                self.model_dir,
            )
            xgboost_models = initialize_xgb_models(
                MODELS[model]["xgboost_model_files"],
                self.model_dir,
                1,
            )
            xgb_vector = np.array(
                ms2pip_pyx.get_vector(peptide, modpeptide, charge), dtype=np.uint16
            )
            xgb_vector = xgb.DMatrix(xgb_vector)
            intensity = []
            for ion_type, model_file in xgboost_models.items():
                preds = model_file.predict(xgb_vector)
                if ion_type in ["x", "y", "y2", "z"]:
                    preds = list(np.array(preds[::-1], dtype=np.float32))
                elif ion_type in ["a", "b", "b2", "c"]:
                    preds = list(np.array(preds, dtype=np.float32))
                intensity.append(preds)
            intensity = np.array(intensity)
        else:
            intensity = np.array(
                ms2pip_pyx.get_predictions(
                    peptide, modpeptide, charge, model_id, peaks_version, ce
                )
            )
        annotation = []
        for ion_type in MODELS[model]["ion_types"]:
            annotation.append(
                [
                    ion_type.lower() + str(i + 1)
                    for i in range(len(mz[MODELS[model]["ion_types"].index(ion_type)]))
                ]
            )
        annotation = np.array(annotation)
        mz = mz.flatten()
        intensity = self._tic_normalize(self._transform(intensity.flatten()))
        annotation = annotation.flatten()
        return mz, intensity, annotation

    def plot_prediction(
        self,
        peptide,
        modifications,
        charge,
        prediction=None,
        ax=None,
        filename=None,
    ):
        """
        Plot MS²PIP-predicted spectrum with spectrum_utils.

        Parameters
        ----------
        peptide: string
            Unmodified peptide sequence. Only canonical amino acids are allowed, and
            peptide sequence should be of length [3, 100].
        modifications: string
            MS²PIP style-formatted modification string (e.g. `0|Acetyl|5|Oxidation`).
            See MS²PIP README.md for more info.
        charge: int
            Peptide precursor charge.
        prediction: tuple or None (default: None)
            Tuple with `ms2pip.single_prediction.SinglePrediction.predict()` output.
        ax: matplotlib.axes.Axes or None (default: None)
            Figure ax to plot onto.
        filename: str or None (default: None)
            Filename to save plot to. File extension defines the format. Figure will
            not be saved if None.

        """
        if not prediction:
            prediction = self.predict(peptide, modifications, charge)
        mz, intensity, annotation = prediction

        identifier = f"{peptide}/{charge}/{modifications}"
        precursor_mz = self.mod_info.calc_precursor_mz(peptide, modifications, charge)
        mod_dict = self._modifications_to_dict(modifications)
        sus_annotation = self._get_sus_annotation(mz, annotation)

        spectrum = sus.MsmsSpectrum(
            identifier,
            precursor_mz,
            charge,
            mz,
            intensity,
            annotation=sus_annotation,
            retention_time=None,
            peptide=peptide,
            modifications=mod_dict,
        )

        if not ax:
            ax = plt.gca()
        sup.spectrum(spectrum, ax=ax)
        ax.set_title("MS²PIP prediction for " + identifier)
        if filename:
            plt.savefig(filename)

    @staticmethod
    def _validate_sequence(sequence):
        """Validate peptide sequence for MS²PIP."""
        pattern = r"[ACDEFGHIKLMNPQRSTVWY]{3,99}"
        if not re.fullmatch(pattern, sequence):
            raise InvalidPeptideError

    @staticmethod
    def _validate_mod_string(mod_string):
        """Validate modification string for MS²PIP."""
        pattern = r"^(?:(?:-1|[0-9]{1,2})\|(?:[^\|\s]+)\|?)*$|^-$"
        if not re.fullmatch(pattern, mod_string):
            raise InvalidModificationFormattingError

    @staticmethod
    def _transform(intensity):
        """Undo MS²PIP peak intensity log transformation and pseudo-count."""
        return (2 ** np.array(intensity)) - 0.001

    @staticmethod
    def _tic_normalize(intensity):
        """TIC-normalize peak intensities."""
        intensity = np.array(intensity)
        return intensity / np.sum(intensity)

    @staticmethod
    def _modifications_to_dict(modifications):
        """Convert ms2pip modification notation to spectrum_utils dict."""

        def parse_loc(loc):
            if loc == "0":
                return "N-term"
            elif loc == "-1":
                return "C-term"
            else:
                return int(loc) - 1

        m_split = [modifications.split("|")[i::2] for i in [0, 1]]
        mods_dict = {parse_loc(loc): name for loc, name in zip(m_split[0], m_split[1])}
        return mods_dict

    @staticmethod
    def _get_sus_annotation(mz, annotation):
        """Get spectrum_utils.PeptideFragmentAnnotation objects."""
        return [
            sus.PeptideFragmentAnnotation(1, mz, annotation[0], annotation[1:])
            for mz, annotation in zip(mz, annotation)
        ]


@click.command()
@click.argument("peptide", type=str)
@click.argument("modifications", type=str)
@click.argument("charge", type=int)
@click.option("-m", "--model", type=str, default="HCD", help="")
@click.option("-c", "--configfile", type=click.Path(exists=True), default=None, help="")
@click.option("--model-dir", type=click.Path(), default=None, help="")
@click.option(
    "-o",
    "--output",
    type=click.Path(exists=False),
    default="ms2pip_prediction.png",
    help="",
)
def _main(
    peptide,
    modifications,
    charge,
    model="HCD",
    configfile=None,
    model_dir=None,
    output="ms2pip_prediction.png",
):
    """
    Generate MS²PIP-predicted spectrum and plot.

    \b
    Examples:
     - ms2pip-single-prediction PGAQANPYSR "-" 3
     - ms2pip-single-prediction -o prediction.png PGAQANPYSR "-" 3
     - ms2pip-single-prediction -c config.toml NSVPCSR "5|Carbamidomethyl" 3

    """
    root_logger = logging.getLogger()
    handler = logging.StreamHandler()
    root_logger.addHandler(handler)
    root_logger.setLevel(logging.INFO)

    if configfile:
        config_parser = ConfigParser(configfile)
        mod_strings = config_parser.config["ms2pip"]["ptm"]
    else:
        mod_strings = None
    ms2pip_sp = SinglePrediction(mod_strings, model_dir)

    _, ax = plt.subplots(figsize=(10, 5))
    prediction = ms2pip_sp.predict(peptide, modifications, charge, model=model)
    ms2pip_sp.plot_prediction(
        peptide,
        modifications,
        charge,
        prediction=prediction,
        ax=ax,
        filename=output,
    )


if __name__ == "__main__":
    _main()



import bisect
import logging
from operator import itemgetter

import numpy as np
import pyteomics.mgf

logger = logging.getLogger("ms2pip.match_spectra")


def get_intense_mzs(mzs, intensity, n=3):
    return [x[0] for x in sorted(zip(mzs, intensity), key=itemgetter(1), reverse=True)[:n]]


def match_mzs(mzs, predicted, max_error=0.02):
    current = 0
    for pred in predicted:
        current = bisect.bisect_right(mzs, pred - max_error, lo=current)
        if current >= len(mzs) or mzs[current] >= pred + max_error:
            return False
    return current < len(mzs)


def get_predicted_peaks(pepids, mzs, intensities):
    # NOTE: we need to concatenate the predictions for b and y ions
    return dict(zip(pepids,
                    (sorted(get_intense_mzs(np.concatenate(_mzs, axis=None),
                                            np.concatenate(_intensities, axis=None)))
                     for _mzs, _intensities in zip(mzs, intensities))))


class MatchSpectra:
    DATA_COLS = ['spec_id', 'peptide', 'modifications', 'charge']

    def __init__(self, peprec, mods, pepids, predicted_mzs, predicted_intensities):
        """
        Initialise spectra matcher

        Parameters:
        -----------
        peprec: pandas.DataFrame
            PEPREC formatted input peptides
        mods: ms2pip.peptides.Modifications
            Modifications used in PEPREC file
        pepids: Iterable[str]
            Iterable of spec_id's from peprec input ordered matching the predictions
        predicted_mzs: Iterable[list[numpy.array[float32]]]
            Iterable of predicted m/z values
        predicted_mzs: Iterable[list[numpy.array[float32]]]
            Iterable of predicted intensities
        """
        self.peprec = peprec
        self.mods = mods
        self.predictions = get_predicted_peaks(pepids, predicted_mzs, predicted_intensities)
        self._generate_peptide_list()

    def _generate_peptide_list(self):
        peptides = [
            (
                spec_id,
                self.mods.calc_precursor_mz(peptide,
                                            modifications,
                                            charge)[1],
                self.predictions[spec_id]
            ) for spec_id, peptide, modifications, charge in self.peprec[self.DATA_COLS].values
        ]
        peptides.sort(key=itemgetter(1))
        self.peptides = peptides

    def match_mgfs(self, mgf_files, max_error=0.02):
        """
        Match predicted spectra to spectra in given MGF files.

        Paramters:
        ----------
        mgf_files: list[str]
            List of filenames of MGF files
        max_error: float, optional
            Maximum error for masses
        """
        logger.info("match predicted spectra to spectra in mgf files (%s)", mgf_files)
        precursors = [x[1] for x in self.peptides]

        for mgf_file in mgf_files:
            logger.debug("open %s", mgf_file)
            with pyteomics.mgf.read(mgf_file, use_header=False, convert_arrays=0, read_charges=False) as reader:
                for spectrum in reader:
                    if 'pepmass' not in spectrum['params']:
                        continue
                    pepmass = spectrum['params']['pepmass'][0]

                    # compare all peptides with a similar precursor m/z
                    i = bisect.bisect_right(precursors, pepmass - max_error)
                    while i < len(precursors) and precursors[i] < pepmass + max_error:
                        spec_id, _, pred_peaks = self.peptides[i]
                        if match_mzs(sorted(spectrum['m/z array']), pred_peaks, max_error=max_error):
                            yield spec_id, mgf_file, spectrum
                        i += 1

    def match_sqldb(self, sqldb_uri="postgresql:///ms2pip", max_error=0.02):
        """
        Match predicted spectra to given database of observed spectra.

        Paramters:
        ----------
        sqldb_uri: str
            URI of database to connect to
        max_error: float, optional
            Maximum error for masses
        """
        from ms2pip.sqldb import tables
        from sqlalchemy import select

        engine = tables.create_engine(sqldb_uri)
        precursors = np.fromiter((x[1] for x in self.peptides), dtype=np.float, count=len(self.peprec))
        gaps = np.where(np.diff(precursors) >= max_error)[0]

        with engine.connect() as connection:
            start = 0
            for end in gaps:
                for spec in connection.execute(
                        select([tables.spec, tables.specfile.c.filename])
                        .select_from(tables.spec.join(tables.specfile)).where(
                            tables.spec.c.pepmass > self.peptides[start][1] - max_error
                        ).where(
                            tables.spec.c.pepmass < self.peptides[end][1] + max_error
                        ).order_by(tables.spec.c.pepmass)):
                    for spec_id, mz, pred_peaks in self.peptides[start:end+1]:
                        if mz > spec.pepmass + max_error:
                            break
                        if mz < spec.pepmass - max_error:
                            start += 1
                            continue
                        if match_mzs(spec.mzs, pred_peaks, max_error=max_error):
                            yield (spec_id,
                                   spec.filename,
                                   {'params': {'title': spec.spec_id},
                                    'm/z array': spec.mzs})
                start = end + 1



#!/usr/bin/env python
import csv
import glob
import itertools
import logging
import multiprocessing
import multiprocessing.dummy
import os
import re
from random import shuffle

import numpy as np
import pandas as pd
import xgboost as xgb
from rich.progress import track

from ms2pip.cython_modules import ms2pip_pyx
from ms2pip.exceptions import (FragmentationModelRequiredError,
                               InvalidModificationFormattingError,
                               InvalidPEPRECError, MissingConfigurationError,
                               NoMatchingSpectraFound,
                               NoValidPeptideSequencesError, TitlePatternError,
                               UnknownFragmentationMethodError,
                               UnknownModificationError,
                               UnknownOutputFormatError)
from ms2pip.feature_names import get_feature_names_new
from ms2pip.match_spectra import MatchSpectra
from ms2pip.ms2pip_tools import calc_correlations, spectrum_output
from ms2pip.peptides import (AMINO_ACID_IDS, Modifications,
                             write_amino_acid_masses)
from ms2pip.predict_xgboost import (get_predictions_xgb,
                                    validate_requested_xgb_model)
from ms2pip.retention_time import RetentionTime
from ms2pip.spectrum import read_spectrum_file

logger = logging.getLogger(__name__)

# Supported output formats
SUPPORTED_OUT_FORMATS = ["csv", "mgf", "msp", "bibliospec", "spectronaut", "dlib"]

# Models and their properties
# id is passed to get_predictions to select model
# ion_types is required to write the ion types in the headers of the result files
# features_version is required to select the features version
MODELS = {
    "CID": {
        "id": 0,
        "ion_types": ["B", "Y"],
        "peaks_version": "general",
        "features_version": "normal",
        "xgboost_model_files": {
            "b": "model_20190107_CID_train_B.xgboost",
            "y": "model_20190107_CID_train_Y.xgboost",
        },
        "model_hash": {
            "model_20190107_CID_train_B.xgboost": "4398c6ebe23e2f37c0aca42b095053ecea6fb427",
            "model_20190107_CID_train_Y.xgboost": "e0a9eb37e50da35a949d75807d66fb57e44aca0f",
        },
    },
    "HCD2019": {
        "id": 1,
        "ion_types": ["B", "Y"],
        "peaks_version": "general",
        "features_version": "normal",
    },
    "TTOF5600": {
        "id": 2,
        "ion_types": ["B", "Y"],
        "peaks_version": "general",
        "features_version": "normal",
        "xgboost_model_files": {
            "b": "model_20190107_TTOF5600_train_B.xgboost",
            "y": "model_20190107_TTOF5600_train_Y.xgboost",
        },
        "model_hash": {
            "model_20190107_TTOF5600_train_B.xgboost": "ab2e28dfbc4ee60640253b0b4c127fc272c9d0ed",
            "model_20190107_TTOF5600_train_Y.xgboost": "f8e9ddd8ca78ace06f67460a2fea0d8fa2623452",
        },
    },
    "TMT": {
        "id": 3,
        "ion_types": ["B", "Y"],
        "peaks_version": "general",
        "features_version": "normal",
    },
    "iTRAQ": {
        "id": 4,
        "ion_types": ["B", "Y"],
        "peaks_version": "general",
        "features_version": "normal",
        "xgboost_model_files": {
            "b": "model_20190107_iTRAQ_train_B.xgboost",
            "y": "model_20190107_iTRAQ_train_Y.xgboost",
        },
        "model_hash": {
            "model_20190107_iTRAQ_train_B.xgboost": "b8d94ad329a245210c652a5b35d724d2c74d0d50",
            "model_20190107_iTRAQ_train_Y.xgboost": "56ae87d56fd434b53fcc1d291745cabb7baf463a",
        },
    },
    "iTRAQphospho": {
        "id": 5,
        "ion_types": ["B", "Y"],
        "peaks_version": "general",
        "features_version": "normal",
        "xgboost_model_files": {
            "b": "model_20190107_iTRAQphospho_train_B.xgboost",
            "y": "model_20190107_iTRAQphospho_train_Y.xgboost",
        },
        "model_hash": {
            "model_20190107_iTRAQphospho_train_B.xgboost": "e283b158cc50e219f42f93be624d0d0ac01d6b49",
            "model_20190107_iTRAQphospho_train_Y.xgboost": "261b2e1810a299ed7ebf193ce1fb81a608c07d3b",
        },
    },
    # ETD': {'id': 6, 'ion_types': ['B', 'Y', 'C', 'Z'], 'peaks_version': 'etd', 'features_version': 'normal'},
    "HCDch2": {
        "id": 7,
        "ion_types": ["B", "Y", "B2", "Y2"],
        "peaks_version": "ch2",
        "features_version": "normal",
    },
    "CIDch2": {
        "id": 8,
        "ion_types": ["B", "Y", "B2", "Y2"],
        "peaks_version": "ch2",
        "features_version": "normal",
        "xgboost_model_files": {
            "b": "model_20190107_CID_train_B.xgboost",
            "y": "model_20190107_CID_train_Y.xgboost",
            "b2": "model_20190107_CID_train_B2.xgboost",
            "y2": "model_20190107_CID_train_Y2.xgboost",
        },
        "model_hash": {
            "model_20190107_CID_train_B.xgboost": "4398c6ebe23e2f37c0aca42b095053ecea6fb427",
            "model_20190107_CID_train_Y.xgboost": "e0a9eb37e50da35a949d75807d66fb57e44aca0f",
            "model_20190107_CID_train_B2.xgboost": "602f2fc648890aebbbe2646252ade658af3221a3",
            "model_20190107_CID_train_Y2.xgboost": "4e4ad0f1d4606c17015aae0f74edba69f684d399",
        },
    },
    "HCD2021": {
        "id": 9,
        "ion_types": ["B", "Y"],
        "peaks_version": "general",
        "features_version": "normal",
        "xgboost_model_files": {
            "b": "model_20210416_HCD2021_B.xgboost",
            "y": "model_20210416_HCD2021_Y.xgboost",
        },
        "model_hash": {
            "model_20210416_HCD2021_B.xgboost": "c086c599f618b199bbb36e2411701fb2866b24c8",
            "model_20210416_HCD2021_Y.xgboost": "22a5a137e29e69fa6d4320ed7d701b61cbdc4fcf",
        },
    },
    "Immuno-HCD": {
        "id": 10,
        "ion_types": ["B", "Y"],
        "peaks_version": "general",
        "features_version": "normal",
        "xgboost_model_files": {
            "b": "model_20210316_Immuno_HCD_B.xgboost",
            "y": "model_20210316_Immuno_HCD_Y.xgboost",
        },
        "model_hash": {
            "model_20210316_Immuno_HCD_B.xgboost": "977466d378de2e89c6ae15b4de8f07800d17a7b7",
            "model_20210316_Immuno_HCD_Y.xgboost": "71948e1b9d6c69cb69b9baf84d361a9f80986fea",
        },
    },
    "CID-TMT": {
        "id": 11,
        "ion_types": ["B", "Y"],
        "peaks_version": "general",
        "features_version": "normal",
        "xgboost_model_files": {
            "b": "model_20220104_CID_TMT_B.xgboost",
            "y": "model_20220104_CID_TMT_Y.xgboost",
        },
        "model_hash": {
            "model_20220104_CID_TMT_B.xgboost": "fa834162761a6ae444bb6523c9c1a174b9738389",
            "model_20220104_CID_TMT_Y.xgboost": "299539179ca55d4ac82e9aed6a4e0bd134a9a41e",
        },
    },
}
MODELS["HCD"] = MODELS["HCD2021"]


def process_peptides(worker_num, data, afile, modfile, modfile2, PTMmap, model):
    """
    Function for each worker to process a list of peptides. The models are
    chosen based on model. PTMmap, Ntermmap and Ctermmap determine the
    modifications applied to each peptide sequence. Returns the predicted
    spectra for all the peptides.
    """

    ms2pip_pyx.ms2pip_init(afile, modfile, modfile2)

    # Prepare output variables
    pepid_buf = []
    peplen_buf = []
    charge_buf = []
    mz_buf = []
    target_buf = None
    prediction_buf = []
    vector_buf = []

    # transform pandas dataframe into dictionary for easy access
    if "ce" in data.columns:
        specdict = (
            data[["spec_id", "peptide", "modifications", "charge", "ce"]]
            .set_index("spec_id")
            .to_dict()
        )
        ces = specdict["ce"]
    else:
        specdict = (
            data[["spec_id", "peptide", "modifications", "charge"]]
            .set_index("spec_id")
            .to_dict()
        )
    pepids = data["spec_id"].tolist()
    peptides = specdict["peptide"]
    modifications = specdict["modifications"]
    charges = specdict["charge"]
    del specdict

    # Track progress for only one worker (good approximation of all workers' progress)
    for pepid in track(
        pepids,
        total=len(pepids),
        disable=worker_num != 0,
        transient=True,
        description="",
    ):
        peptide = peptides[pepid]
        peptide = peptide.replace("L", "I")
        mods = modifications[pepid]

        # TODO: Check if 30 is good default CE!
        colen = 30
        if "ce" in data.columns:
            colen = ces[pepid]

        # Peptides longer then 101 lead to "Segmentation fault (core dumped)"
        if len(peptide) > 100:
            continue

        # convert peptide string to integer list to speed up C code
        peptide = np.array(
            [0] + [AMINO_ACID_IDS[x] for x in peptide] + [0], dtype=np.uint16
        )
        modpeptide = apply_mods(peptide, mods, PTMmap)

        pepid_buf.append(pepid)
        peplen = len(peptide) - 2
        peplen_buf.append(peplen)

        ch = charges[pepid]
        charge_buf.append(ch)

        model_id = MODELS[model]["id"]
        peaks_version = MODELS[model]["peaks_version"]

        # get ion mzs
        mzs = ms2pip_pyx.get_mzs(modpeptide, peaks_version)
        mz_buf.append([np.array(m, dtype=np.float32) for m in mzs])

        # If using xgboost model file, get feature vectors to predict outside of MP.
        # Predictions will be added in `_merge_predictions` function.
        if "xgboost_model_files" in MODELS[model].keys():
            vector_buf.append(
                np.array(
                    ms2pip_pyx.get_vector(peptide, modpeptide, ch),
                    dtype=np.uint16,
                )
            )
        else:
            predictions = ms2pip_pyx.get_predictions(
                peptide, modpeptide, ch, model_id, peaks_version, colen
            )
            prediction_buf.append([np.array(p, dtype=np.float32) for p in predictions])

    return (
        pepid_buf,
        peplen_buf,
        charge_buf,
        mz_buf,
        target_buf,
        prediction_buf,
        vector_buf,
    )


def process_spectra(
    worker_num,
    data,
    spec_file,
    vector_file,
    afile,
    modfile,
    modfile2,
    PTMmap,
    model,
    fragerror,
    spectrum_id_pattern,
):
    """
    Function for each worker to process a list of spectra. Each peptide's
    sequence is extracted from the mgf file. Then models are chosen based on
    model. PTMmap, Ntermmap and Ctermmap determine the modifications
    applied to each peptide sequence and the spectrum is predicted. Then either
    the feature vectors are returned, or a DataFrame with the predicted and
    empirical intensities.
    """
    ms2pip_pyx.ms2pip_init(afile, modfile, modfile2)

    model_id = MODELS[model]["id"]
    peaks_version = MODELS[model]["peaks_version"]

    # transform pandas data structure into dictionary for easy access
    if "ce" in data.columns:
        specdict = (
            data[["spec_id", "peptide", "modifications", "ce"]]
            .set_index("spec_id")
            .to_dict()
        )
        ces = specdict["ce"]
    else:
        specdict = (
            data[["spec_id", "peptide", "modifications", "charge"]].set_index("spec_id").to_dict()
        )
    peptides = specdict["peptide"]
    modifications = specdict["modifications"]
    charges = specdict["charge"]

    # cols contains the names of the computed features
    cols_n = get_feature_names_new()
    if "ce" in data.columns:
        cols_n.append("ce")
    # cols_n = get_feature_names_catboost()

    # if vector_file
    dvectors = []
    dtargets = dict()
    psmids = []

    # else
    pepid_buf = []
    peplen_buf = []
    charge_buf = []
    mz_buf = []
    target_buf = []
    prediction_buf = []
    vector_buf = []

    spectrum_id_regex = re.compile(spectrum_id_pattern)

    # Track progress for only one worker (good approximation of all workers' progress)
    for spectrum in track(
        read_spectrum_file(spec_file),
        total=len(peptides),
        disable=worker_num != 0,
        transient=True,
        description="",
    ):
        # Match title with regex
        match = spectrum_id_regex.search(spectrum.title)
        try:
            title = match[1]
        except (TypeError, IndexError):
            raise TitlePatternError(
                "Spectrum title pattern could not be matched to spectrum IDs "
                f"`{spectrum.title}`. "
                " Are you sure that the regex contains a capturing group?"
            )

        if title not in peptides:
            continue

        peptide = peptides[title]
        peptide = peptide.replace("L", "I")
        mods = modifications[title]

        if spectrum.precursor_charge:
            charge = spectrum.precursor_charge
        else:
            charge = charges[title]  # If charge cannot be parsed from MGF

        if "mut" in mods:
            continue

        # Peptides longer then 101 lead to "Segmentation fault (core dumped)"
        if len(peptide) > 100:
            continue

        # convert peptide string to integer list to speed up C code
        peptide = np.array(
            [0] + [AMINO_ACID_IDS[x] for x in peptide] + [0], dtype=np.uint16
        )

        try:
            modpeptide = apply_mods(peptide, mods, PTMmap)
        except UnknownModificationError as e:
            logger.warn("Unknown modification: %s", e)
            continue

        # Spectrum preprocessing:
        # Remove reporter ions and percursor peak, normalize, tranform
        for label_type in ["iTRAQ", "TMT"]:
            if label_type in model:
                spectrum.remove_reporter_ions("iTRAQ")
        # spectrum.remove_precursor()
        spectrum.tic_norm()
        spectrum.log2_transform()

        # TODO: Check if 30 is good default CE!
        # RG: removed `if ce == 0` in get_vector, split up into two functions
        colen = 30
        if "ce" in data.columns:
            try:
                colen = int(float(ces[title]))
            except:
                logger.warn("Could not parse collision energy!")
                continue

        if vector_file:
            # get targets
            targets = ms2pip_pyx.get_targets(
                modpeptide,
                spectrum.msms,
                spectrum.peaks,
                float(fragerror),
                peaks_version,
            )
            psmids.extend([title] * (len(targets[0])))
            if "ce" in data.columns:
                dvectors.append(
                    np.array(
                        ms2pip_pyx.get_vector_ce(
                            peptide, modpeptide, charge, colen
                        ),
                        dtype=np.uint16,
                    )
                )  # SD: added collision energy
            else:
                dvectors.append(
                    np.array(
                        ms2pip_pyx.get_vector(peptide, modpeptide, charge),
                        dtype=np.uint16,
                    )
                )

            # Collecting targets to dict; works for variable number of ion types
            # For C-term ion types (y, y++, z), flip the order of targets,
            # for correct order in vectors DataFrame
            for i, t in enumerate(targets):
                if i in dtargets.keys():
                    if i % 2 == 0:
                        dtargets[i].extend(t)
                    else:
                        dtargets[i].extend(t[::-1])
                else:
                    if i % 2 == 0:
                        dtargets[i] = [t]
                    else:
                        dtargets[i] = [t[::-1]]

        else:
            # Predict the b- and y-ion intensities from the peptide
            pepid_buf.append(title)
            peplen_buf.append(len(peptide) - 2)
            charge_buf.append(charge)

            # get/append ion mzs, targets and predictions
            targets = ms2pip_pyx.get_targets(
                modpeptide,
                spectrum.msms,
                spectrum.peaks,
                float(fragerror),
                peaks_version,
            )
            target_buf.append([np.array(t, dtype=np.float32) for t in targets])
            mzs = ms2pip_pyx.get_mzs(modpeptide, peaks_version)
            mz_buf.append([np.array(m, dtype=np.float32) for m in mzs])

            # If using xgboost model file, get feature vectors to predict outside of MP.
            # Predictions will be added in `_merge_predictions` function.
            if "xgboost_model_files" in MODELS[model].keys():
                vector_buf.append(
                    np.array(
                        ms2pip_pyx.get_vector(peptide, modpeptide, charge),
                        dtype=np.uint16,
                    )
                )
            else:
                predictions = ms2pip_pyx.get_predictions(
                    peptide, modpeptide, charge, model_id, peaks_version, colen
                )
                prediction_buf.append(
                    [np.array(p, dtype=np.float32) for p in predictions]
                )

    # If feature vectors requested, return specific data
    if vector_file:
        if dvectors:
            # If num_cpu > number of spectra, dvectors can be empty
            if len(dvectors) >= 1:
                # Concatenate dvectors into 2D ndarray before making DataFrame to reduce
                # memory usage
                dvectors = np.concatenate(dvectors)
            df = pd.DataFrame(dvectors, dtype=np.uint16, copy=False)
            df.columns = df.columns.astype(str)
        else:
            df = pd.DataFrame()
        return psmids, df, dtargets

    # Else, return general data
    return (
        pepid_buf,
        peplen_buf,
        charge_buf,
        mz_buf,
        target_buf,
        prediction_buf,
        vector_buf,
    )


def prepare_titles(titles, num_cpu):
    """
    Take a list and return a list containing num_cpu smaller lists with the
    spectrum titles/peptides that will be split across the workers
    """
    # titles might be ordered from small to large peptides,
    # shuffling improves parallel speeds
    shuffle(titles)

    split_titles = [
        titles[i * len(titles) // num_cpu : (i + 1) * len(titles) // num_cpu]
        for i in range(num_cpu)
    ]
    logger.debug(
        "{} spectra (~{:.0f} per cpu)".format(
            len(titles), np.mean([len(a) for a in split_titles])
        )
    )

    return split_titles


def apply_mods(peptide, mods, PTMmap):
    """
    Takes a peptide sequence and a set of modifications. Returns the modified
    version of the peptide sequence, c- and n-term modifications. This modified
    version are hard coded in ms2pipfeatures_c.c for now.
    """
    modpeptide = np.array(peptide[:], dtype=np.uint16)

    if mods != "-":
        l = mods.split("|")
        if len(l) % 2 != 0:
            raise InvalidModificationFormattingError(mods)
        for i in range(0, len(l), 2):
            tl = l[i + 1]
            if tl in PTMmap:
                modpeptide[int(l[i])] = PTMmap[tl]
            else:
                raise UnknownModificationError(tl)

    return modpeptide


def peakcount(x):
    c = 0.0
    for i in x:
        if i > -9.95:
            c += 1.0
    return c / len(x)


class MS2PIP:
    def __init__(
        self,
        pep_file,
        spec_file=None,
        spectrum_id_pattern="(.*)",
        vector_file=None,
        num_cpu=1,
        params=None,
        output_filename=None,
        return_results=False,
        limit=None,
        add_retention_time=False,
        compute_correlations=False,
        match_spectra=False,
        sqldb_uri=None,
        model_dir=None,
    ):
        """
        MS²PIP peak intensity predictor.

        Parameters
        ----------
        pep_file : str or pandas.DataFrame
            Path to file or DataFrame with peptide information (see
            https://github.com/compomics/ms2pip_c#peprec-file)
        spec_file : str, optional
            Path to spectrum file with target intensities. Provide for
            prediction evaluation, or in combination with `vector_file` for
            target extraction.
        spectrum_id_pattern : str, optional
            Regular expression pattern to apply to spectrum titles before matching to
            peptide file entries.
        vector_file : str, optional
            Output filepath for training feature vectors. Provide this to
            extract feature vectors from `spec_file`. Requires `spec_file`.
        num_cpu : int, default: 1
            Number of parallel processes for multiprocessing steps.
        params : dict
            Dictonary with `model`, `frag_error` and `modifications`.
        output_filename : str, optional
            Filepath prefix for output files
        return_results : bool, default: False
            Return results after prediction (`MS2PIP.run()`) instead of writing
            to output files.
        limit : int, optional
            Limit to first N peptides in PEPREC file.
        add_retention_time : bool, default: False
            Add retention time predictions with DeepLC.
        compute_correlations : bool, default: False
            Compute correlations between predictions and targets. Requires
            `spec_file`.
        match_spectra : bool, default: False
            Match spectra in `spec_file` or `sqldb_uri` to peptides in
            `pep_file` based on predicted intensities (experimental).
        sqldb_uri : str, optional
            URI to SQL database for `match_spectra` feature.
        model_dir : str, optional
            Custom directory for downloaded XGBoost model files. By
            default, `~/.ms2pip` is used.

        Examples
        --------
        >>> from ms2pip.ms2pipC import MS2PIP
        >>> params = {
        ...     "ms2pip": {
        ...         "ptm": [
        ...             "Oxidation,15.994915,opt,M",
        ...             "Carbamidomethyl,57.021464,opt,C",
        ...             "Acetyl,42.010565,opt,N-term",
        ...         ],
        ...         "frag_method": "HCD",
        ...         "frag_error": 0.02,
        ...         "out": "csv",
        ...         "sptm": [], "gptm": [],
        ...     }
        ... }
        >>> ms2pip = MS2PIP("test.peprec", params=params)
        >>> ms2pip.run()

        """
        self.pep_file = pep_file
        self.vector_file = vector_file
        self.spectrum_id_pattern = spectrum_id_pattern
        self.num_cpu = num_cpu
        self.params = params
        self.return_results = return_results
        self.limit = limit
        self.add_retention_time = add_retention_time
        self.compute_correlations = compute_correlations
        self.match_spectra = match_spectra
        self.sqldb_uri = sqldb_uri
        self.model_dir = model_dir

        self.afile = None
        self.modfile = None
        self.modfile2 = None

        if params is None:
            raise MissingConfigurationError()

        if "model" in self.params["ms2pip"]:
            self.model = self.params["ms2pip"]["model"]
        elif "frag_method" in self.params["ms2pip"]:
            self.model = self.params["ms2pip"]["frag_method"]
        else:
            raise FragmentationModelRequiredError()
        self.fragerror = self.params["ms2pip"]["frag_error"]

        # Validate requested output formats
        if "out" in self.params["ms2pip"]:
            self.out_formats = [
                o.lower().strip() for o in self.params["ms2pip"]["out"].split(",")
            ]
            for o in self.out_formats:
                if o not in SUPPORTED_OUT_FORMATS:
                    raise UnknownOutputFormatError(o)
        else:
            if not return_results:
                logger.info("No output format specified; defaulting to csv")
                self.out_formats = ["csv"]
            else:
                self.out_formats = []

        # Validate model_dir
        if not self.model_dir:
            self.model_dir = os.path.join(os.path.expanduser("~"), ".ms2pip")

        # Validate requested model
        if self.model in MODELS.keys():
            logger.info("Using %s models", self.model)
            if "xgboost_model_files" in MODELS[self.model].keys():
                validate_requested_xgb_model(
                    MODELS[self.model]["xgboost_model_files"],
                    MODELS[self.model]["model_hash"],
                    self.model_dir,
                )
        else:
            raise UnknownFragmentationMethodError(self.model)

        if output_filename is None and not return_results:
            self.output_filename = "{}_{}".format(
                ".".join(pep_file.split(".")[:-1]), self.model
            )
        else:
            self.output_filename = output_filename

        logger.debug(f"Starting workers (num_cpu={self.num_cpu})...")
        if multiprocessing.current_process().daemon:
            logger.warn(
                "MS2PIP is running in a daemon process. Disabling multiprocessing as daemonic processes can't have children."
            )
            self.myPool = multiprocessing.dummy.Pool(1)
        elif self.num_cpu == 1:
            logger.debug("Using dummy multiprocessing pool.")
            self.myPool = multiprocessing.dummy.Pool(1)
        else:
            self.myPool = multiprocessing.Pool(self.num_cpu)

        if self.match_spectra:
            self.spec_file = None
            if self.sqldb_uri:
                self.spec_files = None
            elif os.path.isdir(spec_file):
                self.spec_files = glob.glob("{}/*.mgf".format(spec_file))
            else:
                self.spec_files = [spec_file]
            logger.debug("Using spectrum files %s", self.spec_files)
        else:
            self.spec_file = spec_file
            self.spec_files = None

        self.mods = Modifications()
        for mod_type in ("sptm", "ptm"):
            self.mods.add_from_ms2pip_modstrings(
                self.params["ms2pip"][mod_type], mod_type=mod_type
            )

    def run(self):
        """Run initiated MS2PIP based on class configuration."""
        self.afile = write_amino_acid_masses()
        self.modfile = self.mods.write_modifications_file(mod_type="ptm")
        self.modfile2 = self.mods.write_modifications_file(mod_type="sptm")

        self._read_peptide_information()

        if self.add_retention_time:
            logger.info("Adding retention time predictions")
            rt_predictor = RetentionTime(config=self.params)
            rt_predictor.add_rt_predictions(self.data)

        # Spectrum file mode
        if self.spec_file:
            logger.info("Processing spectra and peptides...")
            results = self._process_spectra()
            # Feature vectors requested
            if self.vector_file:
                self._write_vector_file(results)
            # Predictions (and targets) requested
            else:
                logger.debug("Merging results")
                all_preds = self._merge_predictions(results)
                # Correlations also requested
                if self.compute_correlations:
                    logger.info("Computing correlations")
                    correlations = calc_correlations.calc_correlations(all_preds)
                    logger.info(
                        "Median correlations: \n%s",
                        str(correlations.groupby("ion")["pearsonr"].median()),
                    )
                    if not self.return_results:
                        corr_filename = self.output_filename + "_correlations.csv"
                        logger.info(f"Writing file {corr_filename}")
                        correlations.to_csv(corr_filename, index=True)
                    else:
                        return correlations
                if not self.return_results:
                    pae_filename = self.output_filename + "_pred_and_emp.csv"
                    logger.info(f"Writing file {pae_filename}...")
                    all_preds.to_csv(pae_filename, index=False)
                else:
                    return all_preds

        # Match spectra mode
        elif self.match_spectra:
            results = self._process_peptides()
            matched_spectra = self._match_spectra(results)
            self._write_matched_spectra(matched_spectra)

        # Predictions-only mode
        else:
            logger.info("Processing peptides...")
            results = self._process_peptides()

            logger.debug("Merging results ...")
            all_preds = self._merge_predictions(results)

            if not self.return_results:
                self._write_predictions(all_preds)
            else:
                return all_preds

    def cleanup(self):
        """Cleanup temporary files."""
        if self.afile:
            os.remove(self.afile)
        if self.modfile:
            os.remove(self.modfile)
        if self.modfile2:
            os.remove(self.modfile2)

    def _read_peptide_information(self):
        # read peptide information
        # the file contains the columns: spec_id, modifications, peptide and charge
        if type(self.pep_file) == str:
            with open(self.pep_file, "rt") as f:
                line = f.readline()
                if line[:7] != "spec_id":
                    raise InvalidPEPRECError()
                sep = line[7]
            data = pd.read_csv(
                self.pep_file,
                sep=sep,
                index_col=False,
                dtype={"spec_id": str, "modifications": str},
                nrows=self.limit,
            )
        else:
            data = self.pep_file
        # for some reason the missing values are converted to float otherwise
        data = data.fillna("-")

        # Filter PEPREC for unsupported peptides
        num_pep = len(data)
        data = data[
            ~(data["peptide"].str.contains("B|J|O|U|X|Z"))
            & ~(data["peptide"].str.len() < 3)
            & ~(data["peptide"].str.len() > 99)
        ].copy()
        num_pep_filtered = num_pep - len(data)
        if num_pep_filtered > 0:
            logger.warning(
                f"Removed {num_pep_filtered} unsupported peptide sequences (< 3, > 99 "
                f"amino acids, or containing B, J, O, U, X or Z). Retained "
                f"{len(data)} entries."
            )

        if len(data) == 0:
            raise NoValidPeptideSequencesError()

        self.data = data

    def _execute_in_pool(self, titles, func, args):
        split_titles = prepare_titles(titles, self.num_cpu)
        results = []
        for i in range(self.num_cpu):
            tmp = split_titles[i]
            results.append(
                self.myPool.apply_async(
                    func,
                    args=(i, self.data[self.data.spec_id.isin(tmp)], *args),
                )
            )
        self.myPool.close()
        self.myPool.join()
        return results

    def _process_spectra(self):
        """
        When an mgf file is provided, MS2PIP either saves the feature vectors to
        train models with or writes a file with the predicted spectra next to
        the empirical one.
        """
        titles = self.data["spec_id"].to_list()

        return self._execute_in_pool(
            titles,
            process_spectra,
            (
                self.spec_file,
                self.vector_file,
                self.afile,
                self.modfile,
                self.modfile2,
                self.mods.ptm_ids,
                self.model,
                self.fragerror,
                self.spectrum_id_pattern,
            ),
        )

    def _write_vector_file(self, results):
        all_results = []
        for r in results:
            psmids, df, dtargets = r.get()

            # dtargets is a dict, containing targets for every ion type (keys are int)
            for i, t in dtargets.items():
                df[
                    "targets_{}".format(MODELS[self.model]["ion_types"][i])
                ] = np.concatenate(t, axis=None)
            df["psmid"] = psmids

            all_results.append(df)

        # Only concat DataFrames with content (we get empty ones if more cpu's than peptides)
        all_results = pd.concat([df for df in all_results if len(df) != 0])

        logger.info("Writing vector file %s...", self.vector_file)
        # write result. write format depends on extension:
        ext = self.vector_file.split(".")[-1]
        if ext == "pkl":
            all_results.to_pickle(self.vector_file + ".pkl")
        elif ext == "csv":
            all_results.to_csv(self.vector_file)
        else:
            # "table" is a tag used to read back the .h5
            all_results.to_hdf(self.vector_file, "table")

        return all_results

    def _merge_predictions(self, results):
        pepid_bufs = []
        peplen_bufs = []
        charge_bufs = []
        mz_bufs = []
        target_bufs = []
        prediction_bufs = []
        vector_bufs = []
        for r in results:
            (
                pepid_buf,
                peplen_buf,
                charge_buf,
                mz_buf,
                target_buf,
                prediction_buf,
                vector_buf,
            ) = r.get()
            pepid_bufs.extend(pepid_buf)
            peplen_bufs.extend(peplen_buf)
            charge_bufs.extend(charge_buf)
            mz_bufs.extend(mz_buf)
            if target_buf:
                target_bufs.extend(target_buf)
            if prediction_buf:
                prediction_bufs.extend(prediction_buf)
            if vector_buf:
                vector_bufs.extend(vector_buf)

        # Validate number of results
        if not mz_bufs:
            raise NoMatchingSpectraFound(
                "No spectra matching titles/IDs from PEPREC could be found in "
                "provided spectrum file."
            )
        logger.debug(f"Gathered data for {len(mz_bufs)} peptides/spectra.")

        # If XGBoost model files are used, first predict outside of MP
        # Temporary hack to move XGB prediction step out of MP; ultimately does not
        # make sense to do this in the `_merge_predictions` step...
        if "xgboost_model_files" in MODELS[self.model].keys():
            logger.debug("Converting feature vectors to XGBoost DMatrix...")
            xgb_vector = xgb.DMatrix(np.vstack(vector_bufs))
            num_ions = [l - 1 for l in peplen_bufs]
            prediction_bufs = get_predictions_xgb(
                xgb_vector,
                num_ions,
                MODELS[self.model],
                self.model_dir,
                num_cpu=self.num_cpu,
            )

        # Reconstruct DataFrame
        logger.debug("Constructing DataFrame with results...")
        num_ion_types = len(MODELS[self.model]["ion_types"])
        ions = []
        ionnumbers = []
        charges = []
        pepids = []
        for pi, pl in enumerate(peplen_bufs):
            [
                ions.extend([ion_type] * (pl - 1))
                for ion_type in MODELS[self.model]["ion_types"]
            ]
            ionnumbers.extend([x + 1 for x in range(pl - 1)] * num_ion_types)
            charges.extend([charge_bufs[pi]] * (num_ion_types * (pl - 1)))
            pepids.extend([pepid_bufs[pi]] * (num_ion_types * (pl - 1)))
        all_preds = pd.DataFrame()
        all_preds["spec_id"] = pepids
        all_preds["charge"] = charges
        all_preds["ion"] = ions
        all_preds["ionnumber"] = ionnumbers
        all_preds["mz"] = np.concatenate(mz_bufs, axis=None)
        all_preds["prediction"] = np.concatenate(prediction_bufs, axis=None)
        if target_bufs:
            all_preds["target"] = np.concatenate(target_bufs, axis=None)
        if "rt" in self.data:
            all_preds = all_preds.merge(
                self.data[["spec_id", "rt"]], on="spec_id", copy=False
            )

        return all_preds

    def _process_peptides(self):
        titles = self.data.spec_id.tolist()
        return self._execute_in_pool(
            titles,
            process_peptides,
            (self.afile, self.modfile, self.modfile2, self.mods.ptm_ids, self.model),
        )

    def _write_predictions(self, all_preds):
        spec_out = spectrum_output.SpectrumOutput(
            all_preds,
            self.data,
            self.params["ms2pip"],
            output_filename=self.output_filename,
        )
        spec_out.write_results(self.out_formats)

    def _match_spectra(self, results):
        mz_bufs, prediction_bufs, _, _, _, pepid_bufs = zip(*(r.get() for r in results))

        match_spectra = MatchSpectra(
            self.data,
            self.mods,
            itertools.chain.from_iterable(pepid_bufs),
            itertools.chain.from_iterable(mz_bufs),
            itertools.chain.from_iterable(prediction_bufs),
        )
        if self.spec_files:
            return match_spectra.match_mgfs(self.spec_files)
        elif self.sqldb_uri:
            return match_spectra.match_sqldb(self.sqldb_uri)
        else:
            raise NotImplementedError

    def _write_matched_spectra(self, matched_spectra):
        filename = f"{self.output_filename}_matched_spectra.csv"
        logger.info("Writing file %s...", filename)

        with open(filename, mode="w") as csv_file:
            csv_writer = csv.writer(csv_file)
            csv_writer.writerow(("spec_id", "matched_file" "matched_title"))
            for pep, spec_file, spec in matched_spectra:
                csv_writer.writerow((pep, spec_file, spec["params"]["title"]))



import argparse
import logging
import multiprocessing
import sys

from rich.console import Console
from rich.logging import RichHandler

from ms2pip.config_parser import ConfigParser
from ms2pip.exceptions import (EmptySpectrumError,
                               FragmentationModelRequiredError,
                               InvalidModificationFormattingError,
                               InvalidPEPRECError, InvalidXGBoostModelError,
                               NoValidPeptideSequencesError,
                               UnknownFragmentationMethodError,
                               UnknownModificationError,
                               UnknownOutputFormatError)
from ms2pip.ms2pipC import MODELS, MS2PIP, SUPPORTED_OUT_FORMATS


def print_logo():
    logo = r"""
 __  __ ___  __ ___ ___ ___
|  \/  / __||_ ) _ \_ _| _ \
| |\/| \__ \/__|  _/| ||  _/
|_|  |_|___/   |_| |___|_|

by CompOmics
sven.degroeve@ugent.be
ralf.gabriels@ugent.be

http://compomics.github.io/projects/ms2pip_c.html
    """
    print(logo)


def argument_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("pep_file", metavar="<PEPREC file>", help="list of peptides")
    parser.add_argument(
        "-c",
        "--config-file",
        metavar="CONFIG_FILE",
        action="store",
        required=True,
        dest="config_file",
        help="Configuration file: text-based (extensions `.txt`, `.config`, or `.ms2pip`) or TOML (extension `.toml`).",
    )
    parser.add_argument(
        "-s",
        "--spectrum-file",
        metavar="SPECTRUM_FILE",
        action="store",
        dest="spec_file",
        help="MGF or mzML spectrum file (optional)",
    )
    parser.add_argument(
        "-w",
        "--vector-file",
        metavar="FEATURE_VECTOR_OUTPUT",
        action="store",
        dest="vector_file",
        help="write feature vectors to FILE.{pkl,h5} (optional)",
    )
    parser.add_argument(
        "-r",
        "--retention-time",
        action="store_true",
        default=False,
        dest="add_retention_time",
        help="add retention time predictions (requires DeepLC python package)",
    )
    parser.add_argument(
        "-x",
        "--correlations",
        action="store_true",
        default=False,
        dest="correlations",
        help="calculate correlations (if spectrum file is given)",
    )
    parser.add_argument(
        "-m",
        "--match-spectra",
        action="store_true",
        default=False,
        dest="match_spectra",
        help="match peptides to spectra based on predicted spectra (if spectrum file is given)",
    )
    parser.add_argument(
        "-n",
        "--num-cpu",
        metavar="NUM_CPU",
        action="store",
        dest="num_cpu",
        type=int,
        help="number of CPUs to use (default: all available)",
    )
    parser.add_argument(
        "--sqldb-uri",
        action="store",
        dest="sqldb_uri",
        help="use sql database of observed spectra instead of spectrum files",
    )
    parser.add_argument(
        "--model-dir",
        action="store",
        dest="model_dir",
        help="Custom directory for downloaded XGBoost model files, default: `~/.ms2pip`",
    )
    args = parser.parse_args()

    if not args.num_cpu:
        args.num_cpu = multiprocessing.cpu_count()

    return args


def main():
    logging.basicConfig(
        format="%(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        level=logging.DEBUG,
        handlers=[RichHandler(
            rich_tracebacks=True, console=Console(), show_level=True, show_path=False
        )],
    )
    logger = logging.getLogger(__name__)

    print_logo()

    args = argument_parser()
    config_parser = ConfigParser(filepath=args.config_file)

    try:
        ms2pip = MS2PIP(
            args.pep_file,
            spec_file=args.spec_file,
            vector_file=args.vector_file,
            params=config_parser.config,
            num_cpu=args.num_cpu,
            add_retention_time=args.add_retention_time,
            compute_correlations=args.correlations,
            match_spectra=args.match_spectra,
            sqldb_uri=args.sqldb_uri,
            model_dir=args.model_dir,
        )
        try:
            ms2pip.run()
        finally:
            ms2pip.cleanup()
    except InvalidPEPRECError:
        logger.critical("PEPREC file should start with header column")
        sys.exit(1)
    except NoValidPeptideSequencesError:
        logger.critical(
            "No peptides for which to predict intensities. \
            please provide at least one valid peptide sequence."
        )
        sys.exit(1)
    except UnknownModificationError as e:
        logger.critical("Unknown modification: %s", e)
        sys.exit(1)
    except InvalidModificationFormattingError as e:
        logger.critical("Invalid formatting of modifications: %s", e)
        sys.exit(1)
    except UnknownOutputFormatError as o:
        logger.critical(
            f"Unknown output format: `{o}` (supported formats: `{SUPPORTED_OUT_FORMATS}`)"
        )
        sys.exit(1)
    except UnknownFragmentationMethodError as f:
        logger.critical(
            f"Unknown model: `{f}` (supported models: {MODELS.keys()})"
        )
        sys.exit(1)
    except FragmentationModelRequiredError:
        logger.critical("Please specify model in config file.")
        sys.exit(1)
    except InvalidXGBoostModelError:
        logger.critical(
            f"Could not download XGBoost model properly\nTry manual download"
        )
        sys.exit(1)
    except EmptySpectrumError:
        logger.critical("Provided MGF file cannot contain empty spectra")
        sys.exit(1)


if __name__ == "__main__":
    main()



"""
Write spectrum files from MS2PIP predictions.
"""

import itertools
import logging
import os
from ast import literal_eval
from functools import wraps
from io import StringIO
from operator import itemgetter
from time import localtime, strftime
from typing import Any, Dict, List

from ms2pip.peptides import Modifications

logger = logging.getLogger("ms2pip.spectrum_output")


class InvalidWriteModeError(ValueError):
    pass


# Writer decorator
def writer(**kwargs):
    def deco(write_function):
        @wraps(write_function)
        def wrapper(self):
            return self._write_general(write_function, **kwargs)

        return wrapper

    return deco


def output_format(output_format):
    class OutputFormat:
        def __init__(self, fn):
            self.fn = fn
            self.output_format = output_format

        def __set_name__(self, owner, name):
            owner.OUTPUT_FORMATS[self.output_format] = self.fn
            setattr(owner, name, self.fn)

    return OutputFormat


class SpectrumOutput:
    """
    Write MS2PIP predictions to various output formats.

    Parameters
    ----------
    all_preds: pd.DataFrame
        MS2PIP predictions
    peprec: pd.DataFrame
        PEPREC with peptide information
    params: dict
        MS2PIP parameters
    output_filename: str, optional
        path and name for output files, will be suffexed with `_predictions` and the
        relevant file extension (default: ms2pip_predictions)
    write_mode: str, optional
        write mode to use: "wt+" to append to start a new file, "at" to append to an
        existing file (default: "wt+")
    return_stringbuffer: bool, optional
        If True, files are written to a StringIO object, which the write function
        returns. If False, files are written to a file on disk.
    is_log_space: bool, optional
        Set to true if predicted intensities in `all_preds` are in log-space. In that
        case, intensities will first be transformed to "normal"-space.

    Methods
    -------
    write_msp()
        Write predictions to MSP file
    write_mgf()
        Write predictions to MGF file
    write_bibliospec()
        Write predictions to Bibliospec SSL/MS2 files (also for Skyline)
    write_spectronaut()
        Write predictions to Spectronaut CSV file
    write_dlib()
        Write predictions to a DLIB SQLite file
    write_csv()
        Write predictions to CSV file
    write_results(output_formats)
        Write MS2PIP predictions in output formats defined by output_formats.

    Example
    -------
    >>> so = ms2pip.spectrum_tools.spectrum_output.SpectrumOutput(
            all_preds,
            peprec,
            params
        )
    >>> so.write_msp()
    >>> so.write_spectronaut()

    """

    OUTPUT_FORMATS = {}

    def __init__(
        self,
        all_preds,
        peprec,
        params,
        output_filename="ms2pip_predictions",
        write_mode="wt+",
        return_stringbuffer=False,
        is_log_space=True,
    ):
        self.all_preds = all_preds
        self.peprec = peprec
        self.params = params
        self.output_filename = output_filename
        self.write_mode = write_mode
        self.return_stringbuffer = return_stringbuffer
        self.is_log_space = is_log_space

        self.peprec_dict = None
        self.preds_dict = None
        self.normalization = None
        self.diff_modification_mapping = {}

        self.has_rt = "rt" in self.peprec.columns
        self.has_protein_list = "protein_list" in self.peprec.columns

        self.mods = Modifications()
        self.mods.add_from_ms2pip_modstrings(params["ptm"])

        if self.write_mode not in ["wt+", "wt", "at", "w", "a"]:
            raise InvalidWriteModeError(self.write_mode)

        if "a" in self.write_mode and self.return_stringbuffer:
            raise InvalidWriteModeError(self.write_mode)

    def _generate_peprec_dict(self, rt_to_seconds=True):
        """
        Create easy to access dict from all_preds and peprec dataframes
        """
        peprec_tmp = self.peprec.copy()

        if self.has_rt and rt_to_seconds:
            peprec_tmp["rt"] = peprec_tmp["rt"] * 60

        peprec_tmp.index = peprec_tmp["spec_id"]
        peprec_tmp.drop("spec_id", axis=1, inplace=True)

        self.peprec_dict = peprec_tmp.to_dict(orient="index")

    def _generate_preds_dict(self):
        """
        Create easy to access dict from peprec dataframes
        """
        self.preds_dict = {}
        preds_list = self.all_preds[
            ["spec_id", "charge", "ion", "ionnumber", "mz", "prediction"]
        ].values.tolist()

        for row in preds_list:
            spec_id = row[0]
            if spec_id in self.preds_dict.keys():
                if row[2] in self.preds_dict[spec_id]["peaks"]:
                    self.preds_dict[spec_id]["peaks"][row[2]].append(tuple(row[3:]))
                else:
                    self.preds_dict[spec_id]["peaks"][row[2]] = [tuple(row[3:])]
            else:
                self.preds_dict[spec_id] = {
                    "charge": row[1],
                    "peaks": {row[2]: [tuple(row[3:])]},
                }

    def _normalize_spectra(self, method="basepeak_10000"):
        """
        Normalize spectra
        """
        if self.is_log_space:
            self.all_preds["prediction"] = (
                (2 ** self.all_preds["prediction"]) - 0.001
            ).clip(lower=0)
            self.is_log_space = False

        if method == "basepeak_10000":
            if self.normalization == "basepeak_10000":
                pass
            elif self.normalization == "basepeak_1":
                self.all_preds["prediction"] *= 10000
            else:
                self.all_preds["prediction"] = self.all_preds.groupby(["spec_id"])[
                    "prediction"
                ].apply(lambda x: (x / x.max()) * 10000)
            self.normalization = "basepeak_10000"

        elif method == "basepeak_1":
            if self.normalization == "basepeak_1":
                pass
            elif self.normalization == "basepeak_10000":
                self.all_preds["prediction"] /= 10000
            else:
                self.all_preds["prediction"] = self.all_preds.groupby(["spec_id"])[
                    "prediction"
                ].apply(lambda x: (x / x.max()))
            self.normalization = "basepeak_1"

        elif method == "tic":
            if self.normalization != "tic":
                self.all_preds["prediction"] = self.all_preds.groupby(["spec_id"])[
                    "prediction"
                ].apply(lambda x: x / x.sum())
            self.normalization = "tic"

        else:
            raise NotImplementedError

    def _get_peak_string(
        self,
        peak_dict,
        sep="\t",
        include_zero=False,
        include_annotations=True,
        intensity_type=float,
    ):
        """
        Get MGF/MSP-like peaklist string
        """
        all_peaks = []
        for ion_type, peaks in peak_dict.items():
            for peak in peaks:
                if not include_zero and peak[2] == 0:
                    continue
                if include_annotations:
                    all_peaks.append(
                        (
                            peak[1],
                            f'{peak[1]:.6f}{sep}{intensity_type(peak[2])}{sep}"{ion_type.lower()}{peak[0]}/0.0"',
                        )
                    )
                else:
                    all_peaks.append((peak[1], f"{peak[1]:.6f}{sep}{peak[2]}"))

        all_peaks = sorted(all_peaks, key=itemgetter(0))
        peak_string = "\n".join([peak[1] for peak in all_peaks])

        return peak_string

    def _get_msp_modifications(self, sequence, modifications):
        """
        Format modifications in MSP-style, e.g. "1/0,E,Glu->pyro-Glu"
        """

        if isinstance(modifications, str):
            if modifications == "-":
                msp_modifications = "0"
            else:
                mods = modifications.split("|")
                mods = [(int(mods[i]), mods[i + 1]) for i in range(0, len(mods), 2)]
                mods = [(x, y) if x == 0 else (x - 1, y) for (x, y) in mods]
                mods = sorted(mods)
                mods = [(str(x), sequence[x], y) for (x, y) in mods]
                msp_modifications = "/".join([",".join(list(x)) for x in mods])
                msp_modifications = f"{len(mods)}/{msp_modifications}"
        else:
            msp_modifications = "0"

        return msp_modifications

    def _parse_protein_string(self, protein_list):
        """
        Parse protein string from list, list string literal, or string.
        """
        if isinstance(protein_list, list):
            protein_string = "/".join(protein_list)
        elif isinstance(protein_list, str):
            try:
                protein_string = "/".join(literal_eval(protein_list))
            except ValueError:
                protein_string = protein_list
        else:
            protein_string = ""
        return protein_string

    def _get_last_ssl_scannr(self):
        """
        Return scan number of last line in a Bibliospec SSL file.
        """
        ssl_filename = "{}_predictions.ssl".format(self.output_filename)
        with open(ssl_filename, "rt") as ssl:
            for line in ssl:
                last_line = line
            last_scannr = int(last_line.split("\t")[1])
        return last_scannr

    def _generate_diff_modification_mapping(self, precision):
        """
        Make modification name -> ssl modification name mapping.
        """
        self.diff_modification_mapping[precision] = {
            ptm.split(",")[0]: "{0:+.{1}f}".format(float(ptm.split(",")[1]), precision)
            for ptm in self.params["ptm"]
        }

    def _get_diff_modified_sequence(self, sequence, modifications, precision=1):
        """
        Build BiblioSpec SSL modified sequence string.
        """
        pep = list(sequence)
        mapping = self.diff_modification_mapping[precision]

        for loc, name in zip(
            modifications.split("|")[::2], modifications.split("|")[1::2]
        ):
            # C-term mod
            if loc == "-1":
                pep[-1] = pep[-1] + "[{}]".format(mapping[name])
            # N-term mod
            elif loc == "0":
                pep[0] = pep[0] + "[{}]".format(mapping[name])
            # Normal mod
            else:
                pep[int(loc) - 1] = pep[int(loc) - 1] + "[{}]".format(mapping[name])
        return "".join(pep)

    @output_format("msp")
    @writer(
        file_suffix="_predictions.msp",
        normalization_method="basepeak_10000",
        requires_dicts=True,
        requires_diff_modifications=False,
    )
    def write_msp(self, file_object):
        """
        Construct MSP string and write to file_object.
        """

        for spec_id in sorted(self.peprec_dict.keys()):
            seq = self.peprec_dict[spec_id]["peptide"]
            mods = self.peprec_dict[spec_id]["modifications"]
            charge = self.peprec_dict[spec_id]["charge"]
            prec_mass, prec_mz = self.mods.calc_precursor_mz(seq, mods, charge)
            msp_modifications = self._get_msp_modifications(seq, mods)
            num_peaks = sum(
                [
                    len(peaklist)
                    for _, peaklist in self.preds_dict[spec_id]["peaks"].items()
                ]
            )

            comment_line = f" Mods={msp_modifications} Parent={prec_mz}"

            if self.has_protein_list:
                protein_list = self.peprec_dict[spec_id]["protein_list"]
                protein_string = self._parse_protein_string(protein_list)
                comment_line += f' Protein="{protein_string}"'

            if self.has_rt:
                rt = self.peprec_dict[spec_id]["rt"]
                comment_line += f" RetentionTime={rt}"

            comment_line += f' MS2PIP_ID="{spec_id}"'

            out = [
                f"Name: {seq}/{charge}",
                f"MW: {prec_mass}",
                f"Comment:{comment_line}",
                f"Num peaks: {num_peaks}",
                self._get_peak_string(
                    self.preds_dict[spec_id]["peaks"],
                    sep="\t",
                    include_annotations=True,
                    intensity_type=int,
                ),
            ]

            file_object.writelines([line + "\n" for line in out] + ["\n"])

    @output_format("mgf")
    @writer(
        file_suffix="_predictions.mgf",
        normalization_method="basepeak_10000",
        requires_dicts=True,
        requires_diff_modifications=False,
    )
    def write_mgf(self, file_object):
        """
        Construct MGF string and write to file_object
        """
        for spec_id in sorted(self.peprec_dict.keys()):
            seq = self.peprec_dict[spec_id]["peptide"]
            mods = self.peprec_dict[spec_id]["modifications"]
            charge = self.peprec_dict[spec_id]["charge"]
            prec_mass, prec_mz = self.mods.calc_precursor_mz(seq, mods, charge)
            msp_modifications = self._get_msp_modifications(seq, mods)

            if self.has_protein_list:
                protein_list = self.peprec_dict[spec_id]["protein_list"]
                protein_string = self._parse_protein_string(protein_list)
            else:
                protein_string = ""

            out = [
                "BEGIN IONS",
                f"TITLE={spec_id} {seq}/{charge} {msp_modifications} {protein_string}",
                f"PEPMASS={prec_mz}",
                f"CHARGE={charge}+",
            ]

            if self.has_rt:
                rt = self.peprec_dict[spec_id]["rt"]
                out.append(f"RTINSECONDS={rt}")

            out.append(
                self._get_peak_string(
                    self.preds_dict[spec_id]["peaks"],
                    sep=" ",
                    include_annotations=False,
                )
            )
            out.append("END IONS\n")
            file_object.writelines([line + "\n" for line in out])

    @output_format("spectronaut")
    @writer(
        file_suffix="_predictions_spectronaut.csv",
        normalization_method="tic",
        requires_dicts=False,
        requires_diff_modifications=True,
    )
    def write_spectronaut(self, file_obj):
        """
        Construct spectronaut DataFrame and write to file_object.
        """
        if "w" in self.write_mode:
            header = True
        elif "a" in self.write_mode:
            header = False
        else:
            raise InvalidWriteModeError(self.write_mode)

        spectronaut_peprec = self.peprec.copy()

        # ModifiedPeptide and PrecursorMz columns
        spectronaut_peprec["ModifiedPeptide"] = spectronaut_peprec.apply(
            lambda row: self._get_diff_modified_sequence(
                row["peptide"], row["modifications"]
            ),
            axis=1,
        )
        spectronaut_peprec["PrecursorMz"] = spectronaut_peprec.apply(
            lambda row: self.mods.calc_precursor_mz(
                row["peptide"], row["modifications"], row["charge"]
            )[1],
            axis=1,
        )
        spectronaut_peprec["ModifiedPeptide"] = (
            "_" + spectronaut_peprec["ModifiedPeptide"] + "_"
        )

        # Additional columns
        spectronaut_peprec["FragmentLossType"] = "noloss"

        # Retention time
        if "rt" in spectronaut_peprec.columns:
            rt_cols = ["iRT"]
            spectronaut_peprec["iRT"] = spectronaut_peprec["rt"]
        else:
            rt_cols = []

        # ProteinId
        if self.has_protein_list:
            spectronaut_peprec["ProteinId"] = spectronaut_peprec["protein_list"].apply(
                self._parse_protein_string
            )
        else:
            spectronaut_peprec["ProteinId"] = spectronaut_peprec["spec_id"]

        # Rename columns and merge with predictions
        spectronaut_peprec = spectronaut_peprec.rename(
            columns={"charge": "PrecursorCharge", "peptide": "StrippedPeptide"}
        )
        peptide_cols = (
            [
                "ModifiedPeptide",
                "StrippedPeptide",
                "PrecursorCharge",
                "PrecursorMz",
                "ProteinId",
            ]
            + rt_cols
            + ["FragmentLossType"]
        )
        spectronaut_df = spectronaut_peprec[peptide_cols + ["spec_id"]]
        spectronaut_df = self.all_preds.merge(spectronaut_df, on="spec_id")

        # Fragment columns
        spectronaut_df["FragmentCharge"] = (
            spectronaut_df["ion"].str.contains("2").map({True: 2, False: 1})
        )
        spectronaut_df["FragmentType"] = spectronaut_df["ion"].str[0].str.lower()

        # Rename and sort columns
        spectronaut_df = spectronaut_df.rename(
            columns={
                "mz": "FragmentMz",
                "prediction": "RelativeIntensity",
                "ionnumber": "FragmentNumber",
            }
        )
        fragment_cols = [
            "FragmentCharge",
            "FragmentMz",
            "RelativeIntensity",
            "FragmentType",
            "FragmentNumber",
        ]
        spectronaut_df = spectronaut_df[peptide_cols + fragment_cols]
        spectronaut_df.to_csv(file_obj, index=False, header=header)

        return file_obj

    def _write_bibliospec_core(self, file_obj_ssl, file_obj_ms2, start_scannr=0):
        """
        Construct Bibliospec SSL/MS2 strings and write to file_objects.
        """

        for i, spec_id in enumerate(sorted(self.preds_dict.keys())):
            scannr = i + start_scannr
            seq = self.peprec_dict[spec_id]["peptide"]
            mods = self.peprec_dict[spec_id]["modifications"]
            charge = self.peprec_dict[spec_id]["charge"]
            prec_mass, prec_mz = self.mods.calc_precursor_mz(seq, mods, charge)
            ms2_filename = os.path.basename(self.output_filename) + "_predictions.ms2"

            peaks = self._get_peak_string(
                self.preds_dict[spec_id]["peaks"],
                sep="\t",
                include_annotations=False,
            )

            if isinstance(mods, str) and mods != "-" and mods != "":
                mod_seq = self._get_diff_modified_sequence(seq, mods)
            else:
                mod_seq = seq

            rt = self.peprec_dict[spec_id]["rt"] if self.has_rt else ""

            # TODO: implement csv instead of manual writing
            file_obj_ssl.write(
                "\t".join(
                    [ms2_filename, str(scannr), str(charge), mod_seq, "", "", str(rt)]
                )
                + "\n"
            )
            file_obj_ms2.write(
                "\n".join(
                    [
                        f"S\t{scannr}\t{prec_mz}",
                        f"Z\t{charge}\t{prec_mass}",
                        f"D\tseq\t{seq}",
                        f"D\tmodified seq\t{mod_seq}",
                        peaks,
                    ]
                )
                + "\n"
            )

    def _write_general(
        self,
        write_function,
        file_suffix,
        normalization_method,
        requires_dicts,
        requires_diff_modifications,
        diff_modification_precision=1,
    ):
        """
        General write function to call core write functions.

        Note: Does not work for write_bibliospec and write_dlib functions.
        """

        # Normalize if necessary and make dicts
        if not self.normalization == normalization_method:
            self._normalize_spectra(method=normalization_method)
            if requires_dicts:
                self._generate_preds_dict()
        elif requires_dicts and not self.preds_dict:
            self._generate_preds_dict()
        if requires_dicts and not self.peprec_dict:
            self._generate_peprec_dict()

        if (
            requires_diff_modifications
            and diff_modification_precision not in self.diff_modification_mapping
        ):
            self._generate_diff_modification_mapping(diff_modification_precision)

        # Write to file or stringbuffer
        if self.return_stringbuffer:
            file_object = StringIO()
            logger.info("Writing results to StringIO using %s", write_function.__name__)
        else:
            f_name = self.output_filename + file_suffix
            file_object = open(f_name, self.write_mode)
            logger.info("Writing results to %s", f_name)

        write_function(self, file_object)

        return file_object

    @output_format("bibliospec")
    def write_bibliospec(self):
        """
        Write MS2PIP predictions to BiblioSpec SSL and MS2 spectral library files
        (For example for use in Skyline).
        """

        precision = 1
        if precision not in self.diff_modification_mapping:
            self._generate_diff_modification_mapping(precision)

        # Normalize if necessary and make dicts
        if not self.normalization == "basepeak_10000":
            self._normalize_spectra(method="basepeak_10000")
            self._generate_preds_dict()
        elif not self.preds_dict:
            self._generate_preds_dict()
        if not self.peprec_dict:
            self._generate_peprec_dict()

        if self.return_stringbuffer:
            file_obj_ssl = StringIO()
            file_obj_ms2 = StringIO()
        else:
            file_obj_ssl = open(
                "{}_predictions.ssl".format(self.output_filename), self.write_mode
            )
            file_obj_ms2 = open(
                "{}_predictions.ms2".format(self.output_filename), self.write_mode
            )

        # If a new file is written, write headers
        if "w" in self.write_mode:
            start_scannr = 0
            ssl_header = [
                "file",
                "scan",
                "charge",
                "sequence",
                "score-type",
                "score",
                "retention-time",
                "\n",
            ]
            file_obj_ssl.write("\t".join(ssl_header))
            file_obj_ms2.write(
                "H\tCreationDate\t{}\n".format(
                    strftime("%Y-%m-%d %H:%M:%S", localtime())
                )
            )
            file_obj_ms2.write("H\tExtractor\tMS2PIP predictions\n")
        else:
            # Get last scan number of ssl file, to continue indexing from there
            # because Bibliospec speclib scan numbers can only be integers
            start_scannr = self._get_last_ssl_scannr() + 1

        self._write_bibliospec_core(
            file_obj_ssl, file_obj_ms2, start_scannr=start_scannr
        )

        return file_obj_ssl, file_obj_ms2

    def _write_dlib_metadata(self, connection):
        from ms2pip.ms2pip_tools.dlib import DLIB_VERSION, Metadata
        from sqlalchemy import select

        with connection.begin():
            version = connection.execute(
                select([Metadata.c.Value]).where(Metadata.c.Key == "version")
            ).scalar()
            if version is None:
                connection.execute(
                    Metadata.insert().values(
                        Key="version",
                        Value=DLIB_VERSION,
                    )
                )

    def _write_dlib_entries(self, connection, precision):
        from ms2pip.ms2pip_tools.dlib import Entry

        peptide_to_proteins = set()

        with connection.begin():
            for spec_id, peprec in self.peprec_dict.items():
                seq = peprec["peptide"]
                mods = peprec["modifications"]
                charge = peprec["charge"]

                prec_mass, prec_mz = self.mods.calc_precursor_mz(seq, mods, charge)
                mod_seq = self._get_diff_modified_sequence(
                    seq, mods, precision=precision
                )

                all_peaks = sorted(
                    itertools.chain.from_iterable(
                        self.preds_dict[spec_id]["peaks"].values()
                    ),
                    key=itemgetter(1),
                )
                mzs = [peak[1] for peak in all_peaks]
                intensities = [peak[2] for peak in all_peaks]

                connection.execute(
                    Entry.insert().values(
                        PrecursorMz=prec_mz,
                        PrecursorCharge=charge,
                        PeptideModSeq=mod_seq,
                        PeptideSeq=seq,
                        Copies=1,
                        RTInSeconds=peprec["rt"],
                        Score=0,
                        MassEncodedLength=len(mzs),
                        MassArray=mzs,
                        IntensityEncodedLength=len(intensities),
                        IntensityArray=intensities,
                        SourceFile=self.output_filename,
                    )
                )

                if self.has_protein_list:
                    protein_list = peprec["protein_list"]
                    if isinstance(protein_list, str):
                        protein_list = literal_eval(protein_list)

                    for protein in protein_list:
                        peptide_to_proteins.add((seq, protein))

        return peptide_to_proteins

    def _write_dlib_peptide_to_protein(self, connection, peptide_to_proteins):
        from ms2pip.ms2pip_tools.dlib import PeptideToProtein

        if not self.has_protein_list:
            return

        with connection.begin():
            sql_peptide_to_proteins = set()
            proteins = {protein for _, protein in peptide_to_proteins}
            for peptide_to_protein in connection.execute(
                PeptideToProtein.select().where(
                    PeptideToProtein.c.ProteinAccession.in_(proteins)
                )
            ):
                sql_peptide_to_proteins.add(
                    (
                        peptide_to_protein.PeptideSeq,
                        peptide_to_protein.ProteinAccession,
                    )
                )

            peptide_to_proteins.difference_update(sql_peptide_to_proteins)
            for seq, protein in peptide_to_proteins:
                connection.execute(
                    PeptideToProtein.insert().values(
                        PeptideSeq=seq, isDecoy=False, ProteinAccession=protein
                    )
                )

    @output_format("dlib")
    def write_dlib(self):
        """
        Write MS2PIP predictions to a DLIB SQLite file.
        """
        from ms2pip.ms2pip_tools.dlib import metadata, open_sqlite

        normalization = "basepeak_10000"
        precision = 5
        if not self.normalization == normalization:
            self._normalize_spectra(method=normalization)
            self._generate_preds_dict()
        if not self.peprec_dict:
            self._generate_peprec_dict()
        if precision not in self.diff_modification_mapping:
            self._generate_diff_modification_mapping(precision)

        filename = "{}.dlib".format(self.output_filename)
        logger.info("Writing results to %s", filename)

        logger.debug(
            "write mode is ignored for DLIB at the file mode, although append or not is respected"
        )
        if "a" not in self.write_mode and os.path.exists(filename):
            os.remove(filename)

        if self.return_stringbuffer:
            raise NotImplementedError(
                "`return_stringbuffer` not implemented for DLIB output."
            )

        if not self.has_rt:
            raise NotImplementedError("Retention times required to write DLIB file.")

        with open_sqlite(filename) as connection:
            metadata.create_all()
            self._write_dlib_metadata(connection)
            peptide_to_proteins = self._write_dlib_entries(connection, precision)
            self._write_dlib_peptide_to_protein(connection, peptide_to_proteins)

    def get_normalized_predictions(self, normalization_method="tic"):
        """
        Return normalized copy of predictions.
        """
        self._normalize_spectra(method=normalization_method)
        return self.all_preds.copy()

    @output_format("csv")
    def write_csv(self):
        """
        Write MS2PIP predictions to CSV.
        """

        self._normalize_spectra(method="tic")

        # Write to file or stringbuffer
        if self.return_stringbuffer:
            file_object = StringIO()
            logger.info("Writing results to StringIO using %s", "write_csv")
        else:
            f_name = "{}_predictions.csv".format(self.output_filename)
            file_object = open(f_name, self.write_mode)
            logger.info("Writing results to %s", f_name)

        self.all_preds.to_csv(file_object, float_format="%.6g", index=False)
        return file_object

    def write_results(self, output_formats: List[str]) -> Dict[str, Any]:
        """
        Write MS2PIP predictions in output formats defined by output_formats.
        """
        results = {}
        for output_format in output_formats:
            output_format = output_format.lower()
            writer = self.OUTPUT_FORMATS[output_format]
            results[output_format] = writer(self)
        return results



import platform
from glob import glob

import numpy
from Cython.Distutils import build_ext
from setuptools import setup
from setuptools.extension import Extension

VERSION = "3.10.1"

NAME = "ms2pip"
LICENSE = "apache-2.0"
DESCRIPTION = "MS²PIP: MS² Peak Intensity Prediction"
AUTHOR = (
    "Sven Degroeve, Ralf Gabriels, Kevin Velghe, Ana Sílvia C. Silva, Arthur Declercq"
)
AUTHOR_EMAIL = "sven.degroeve@vib-ugent.be"
URL = "https://www.github.com/compomics/ms2pip_c"
PROJECT_URLS = {
    "Documentation": "http://compomics.github.io/projects/ms2pip_c",
    "Source": "https://github.com/compomics/ms2pip_c",
    "Tracker": "https://github.com/compomics/ms2pip_c/issues",
    "Web server": "https://iomics.ugent.be/ms2pip/",
    "Publication": "https://doi.org/10.1093/nar/gkz299/",
}
KEYWORDS = [
    "MS2PIP",
    "Proteomics",
    "peptides",
    "peak intensity prediction",
    "spectrum",
    "machine learning",
    "spectral library",
    "fasta2speclib",
]
CLASSIFIERS = [
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: Apache Software License",
    "Operating System :: POSIX :: Linux",
    "Programming Language :: Python :: 3 :: Only",
    "Topic :: Scientific/Engineering :: Bio-Informatics",
    "Development Status :: 5 - Production/Stable",
]
INSTALL_REQUIRES = [
    "biopython>=1.74,<2",
    "numpy>=1.16,<2",
    "pandas>=0.24,<2",
    "pyteomics>=3.5,<5",
    "tqdm>=4,<5",
    "tables>=3.4",
    "tomlkit>=0.5,<1",
    "sqlalchemy>=1.3,<2",
    "click>=7,<9",
    "xgboost>=1.3,<2",
    "matplotlib>=3.0",
    "spectrum-utils==0.3.5",
    "lxml>=4",
    "rich>=13",
]
PYTHON_REQUIRES = ">=3.7,<4"

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

to_remove = [
    "ms2pip/cython_modules/ms2pip_pyx.c*",
    "ms2pip/cython_modules/ms2pip_pyx.so",
]
# _ = [[os.remove(f) for f in glob(pat)] for pat in to_remove]

# Large machine-written C model files require optimization to be disabled
compile_args = {
    "Linux": [
        "-O0",
        "-fno-var-tracking",
        "-Wno-unused-result",
        "-Wno-cpp",
        "-Wno-unused-function",
    ],
    "Darwin": [
        "-O0",
    ],
    "Windows": [
        "/Od",
        "/DEBUG",
        "/GL-",
        "/bigobj",
        "/wd4244",
    ],
}

extensions = [
    Extension(
        "ms2pip.cython_modules.ms2pip_pyx",
        sources=["ms2pip/cython_modules/ms2pip_pyx.pyx"] + glob("ms2pip/models/*/*.c"),
        extra_compile_args=compile_args[platform.system()],
    )
]

setup(
    name=NAME,
    version=VERSION,
    license=LICENSE,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    project_urls=PROJECT_URLS,
    keywords=KEYWORDS,
    classifiers=CLASSIFIERS,
    packages=["ms2pip", "ms2pip.ms2pip_tools", "fasta2speclib"],
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "ms2pip=ms2pip.__main__:main",
            "fasta2speclib=fasta2speclib.fasta2speclib:main",
            "ms2pip-single-prediction=ms2pip.single_prediction:_main",
        ],
    },
    install_requires=INSTALL_REQUIRES,
    python_requires=PYTHON_REQUIRES,
    ext_modules=extensions,
    include_dirs=[numpy.get_include()],
    cmdclass={"build_ext": build_ext},
)



import zlib

import numpy
import sqlalchemy
from sqlalchemy import (Boolean, Column, Float, Index, Integer, MetaData,
                        String, Table, TypeDecorator)
from sqlalchemy.dialects.sqlite import BLOB


DLIB_VERSION = "0.1.14"


class CompressedArray(TypeDecorator):
    """ Sqlite-like does not support arrays.
        Let's use a custom type decorator.

        See http://docs.sqlalchemy.org/en/latest/core/types.html#sqlalchemy.types.TypeDecorator
    """
    impl = BLOB

    def __init__(self, dtype, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dtype = dtype

    def process_bind_param(self, value, dialect):
        value = numpy.asarray(value, dtype=self.dtype)
        return zlib.compress(value.tobytes(), 6)

    def process_result_value(self, value, dialect):
        decompressed = zlib.decompress(value)
        return numpy.frombuffer(decompressed, dtype=self.dtype).tolist()

    def copy(self):
        # NOTE: length will be passed through to BLOB
        return CompressedArray(self.dtype, self.impl.length)


metadata = MetaData()

big_float = numpy.dtype('>f4')
big_double = numpy.dtype('>f8')

Entry = Table(
    'entries',
    metadata,
    Column('PrecursorMz', Float, nullable=False, index=True),
    Column('PrecursorCharge', Integer, nullable=False),
    Column('PeptideModSeq', String, nullable=False),
    Column('PeptideSeq', String, nullable=False, index=True),
    Column('Copies', Integer, nullable=False),
    Column('RTInSeconds', Float, nullable=False),
    Column('Score', Float, nullable=False),
    Column('MassEncodedLength', Integer, nullable=False),
    Column('MassArray', CompressedArray(big_double), nullable=False),
    Column('IntensityEncodedLength', Integer, nullable=False),
    Column('IntensityArray', CompressedArray(big_float), nullable=False),
    Column('CorrelationEncodedLength', Integer, nullable=True),
    Column('CorrelationArray', CompressedArray(big_float), nullable=True),
    Column('RTInSecondsStart', Float, nullable=True),
    Column('RTInSecondsStop', Float, nullable=True),
    Column('MedianChromatogramEncodedLength', Integer, nullable=True),
    Column('MedianChromatogramArray', CompressedArray(big_float), nullable=True),
    Column('SourceFile', String, nullable=False),
)

Index('ix_entries_PeptideModSeq_PrecursorCharge_SourceFile', Entry.c.PeptideModSeq, Entry.c.PrecursorCharge, Entry.c.SourceFile)

PeptideToProtein = Table(
    'peptidetoprotein',
    metadata,
    Column('PeptideSeq', String, nullable=False, index=True),
    Column('isDecoy', Boolean, nullable=True),
    Column('ProteinAccession', String, nullable=False, index=True),
)

Metadata = Table(
    'metadata',
    metadata,
    Column('Key', String, nullable=False, index=True),
    Column('Value', String, nullable=False),
)


def open_sqlite(filename):
    engine = sqlalchemy.create_engine(f"sqlite:///{filename}")
    metadata.bind = engine
    return engine.connect()
